from typing import Any, Mapping, Optional
from copy import deepcopy
import importlib
import inspect
import tldextract
from surfari.agents.navigation_agent._typing import (
    LLMActionStep,
    LLMResponse,
    ResolveInput,
    ResolveOutput,
    Resolver,
    ResolverLike
)
import surfari.util.surfari_logger as surfari_logger
from surfari.security.site_credential_manager import SiteCredentialManager

logger = surfari_logger.getLogger(__name__)

def extract_steps(resp: LLMResponse) -> Optional[list[LLMActionStep]]:
    """
    Return a *list of step dicts* from `resp` or None if neither `step` nor `steps`
    is present in a valid form.
    - `step`: dict → [dict], list[dict] → list
    - `steps`: list[dict] → list
    """
    if "step" in resp and resp["step"] is not None:
        s = resp["step"]
        if isinstance(s, dict):
            return [s]  # single step
        if isinstance(s, list):
            return s    # already a list of steps
        return None

    if "steps" in resp and resp["steps"] is not None:
        s = resp["steps"]
        if isinstance(s, list):
            return s
        return None

    return None

def _response_has_resolve_value(resp: LLMResponse) -> bool:
    """
    True if `resp` has a `step` or `steps` branch containing at least one
    step dict with a non-empty string `resolve_value`.
    """
    steps = extract_steps(resp)
    if not steps:
        return False

    for st in steps:
        rv = st.get("resolve_value")
        if isinstance(rv, str) and rv.strip():
            return True
    return False

def _call_resolver(
    resolver: ResolverLike,
    prompt: str,
    context: Optional[Mapping[str, Any]]
) -> Optional[str]:
    if resolver is None:
        return None
    if isinstance(resolver, Resolver):
        out = resolver.resolve(ResolveInput(text=prompt, context=context))
        return out.value if isinstance(out, ResolveOutput) else out
    return resolver(prompt, context)  # type: ignore[misc]

def _resolve_steps(steps: list[LLMActionStep], resolver: ResolverLike, context: Optional[Mapping[str, Any]]) -> None:
    assert resolver is not None, "_resolve_steps called with resolver=None"    
    for step in steps:
        if "value" in step:
            step.pop("resolve_value", None)
            continue
        rv = step.get("resolve_value")
        if isinstance(rv, str) and rv.strip():
            prompt = rv.strip()
            try:
                resolved = _call_resolver(resolver, prompt, context)
            except Exception as e:
                logger.exception("Resolver failed for prompt %r (step target=%r)", prompt, step.get("target"))
                continue
            if resolved:
                step["orig_value"] = prompt
                step["value"] = resolved
                # Always remove resolve_value once processed or skipped
                del step["resolve_value"]

def resolve_missing_value_in_llm_response(
    resp: LLMResponse,
    resolver: ResolverLike,
    *,
    context: Optional[Mapping[str, Any]] = None,
    mutate: bool = False
) -> LLMResponse:
    """
    Resolves `resolve_value` → `value` for whichever branch exists:
      - `step`: dict or list[dict]
      - `steps`: list[dict]

    For each step that has `resolve_value` (and no `value` already):
      - sets `orig_value` to the original prompt
      - sets `value` to the resolved string
      - deletes `resolve_value`
    """
    steps_to_process = extract_steps(resp)
    if not steps_to_process:
        return resp  # nothing to do
    
    for step in steps_to_process:
        if "resolve_value" in step and (step["resolve_value"] == "OTP" or "**" in step["resolve_value"]):
            step["orig_value"] = step["resolve_value"].strip()
            step["value"] = step["resolve_value"].strip()
            del step["resolve_value"]

    has_rv = _response_has_resolve_value(resp)
    if not has_rv:
        return resp  # nothing to do

    out: LLMResponse = resp if mutate else deepcopy(resp)  # type: ignore[assignment]
    steps_to_process = extract_steps(out)

    secret_resolver = SecretResolver((context or {}).get("site_id", 9999))
    _resolve_steps(steps_to_process, secret_resolver, context)
    
    has_rv = _response_has_resolve_value(out)
    if not has_rv:
        return out  # Done resolving secrets

    if resolver:
        _resolve_steps(steps_to_process, resolver, context)
    
    has_rv = _response_has_resolve_value(out)
    if not has_rv:
        return out  # Done resolving values       

    logger.warning("Some values could not be resolved, delegating to user.")
    return DefaultDelegationResolver().delegate_to_user(out, mutate=mutate)

# code to load a custom value resolver
class ResolverLoadError(RuntimeError):
    pass

def _import_obj(target: str) -> Any:
    mod_path, sep, qual = target.partition(":")
    if not sep or not qual:
        raise ResolverLoadError(f"Resolver target must look like 'pkg.mod:Name', got '{target}'")
    try:
        module = importlib.import_module(mod_path)
    except Exception as e:
        raise ResolverLoadError(f"Failed to import module '{mod_path}': {e}") from e
    obj = module
    for part in qual.split("."):
        try:
            obj = getattr(obj, part)
        except AttributeError as e:
            raise ResolverLoadError(f"'{qual}' not found in '{mod_path}': missing '{part}'") from e
    return obj

def _validate_callable_two_args(fn: Any) -> bool:
    if not callable(fn):
        return False
    sig = inspect.signature(fn)
    params = list(sig.parameters.values())
    # We require at least two params (prompt, context). They can be positional-or-keyword.
    return len(params) >= 2

def _as_resolver_like(obj: Any) -> ResolverLike:
    if isinstance(obj, Resolver):
        return obj
    if _validate_callable_two_args(obj):
        return obj  # type: ignore[return-value]
    raise ResolverLoadError(
        "Loaded resolver is neither a Resolver instance nor a callable(prompt, context) -> str."
    )

def _normalize_resolver_cfg(cfg: str | Mapping[str, Any]) -> tuple[str, Mapping[str, Any]]:
    """
    Returns (target, params) from either a string or a mapping config.
    """
    if isinstance(cfg, str):
        return cfg.strip(), {}
    if not isinstance(cfg, Mapping):
        raise ResolverLoadError("Resolver config must be a string or a mapping.")
    target = cfg.get("target")
    if not isinstance(target, str):
        raise ResolverLoadError("Resolver mapping must include a string 'target' key.")
    params = cfg.get("params") or {}
    if not isinstance(params, Mapping):
        raise ResolverLoadError("'params' must be a mapping if provided.")
    return target.strip(), params


def _build_resolver_from_target(target: str, params: Mapping[str, Any]) -> ResolverLike:
    """
    Handles alias, class, factory, or callable.
    - If class: instantiate with **params.
    - If callable: try factory(**params) first; if that fails and params is empty,
      try zero-arg factory(); otherwise treat it as the final callable.
    """
    logger.info(f"Building resolver from target: {target}, params: {params}")

    obj = _import_obj(target)

    # Class -> instance
    if inspect.isclass(obj):
        try:
            instance = obj(**params)
        except TypeError as e:
            raise ResolverLoadError(f"Failed to construct resolver class: {e}") from e
        return _as_resolver_like(instance)

    # Callable -> factory or final callable
    if callable(obj):
        # Prefer factory(**params)
        try:
            produced = obj(**params)
            if produced is not None:
                return _as_resolver_like(produced)
        except TypeError:
            # If no params provided, try zero-arg factory()
            if not params:
                try:
                    produced = obj()
                    if produced is not None:
                        return _as_resolver_like(produced)
                except TypeError:
                    pass
        # Otherwise, treat the callable itself as the resolver
        return _as_resolver_like(obj)

    raise ResolverLoadError(f"Unsupported resolver object: {type(obj)!r}")


def create_resolver_from_config(cfg: str | Mapping[str, Any]) -> ResolverLike:
    """
    cfg can be:
      - "pkg.mod:ClassName"
      - "pkg.mod:function_name" (callable(prompt, context) -> str)
      - {"target": "...", "params": {...}}  # for class __init__ or factory(**params)
    """
    target, params = _normalize_resolver_cfg(cfg)
    return _build_resolver_from_target(target, params)

def base_domains_match(url1: str, url2: str) -> bool:
    """
    Returns True if the base registrable domains match for the two URLs.
    For example:
        login.sbc.com  -> sbc.com
        www.sbc.com    -> sbc.com
        sub.level.sbc.com/hello?jsp -> sbc.com
    """
    def get_base_domain(url: str) -> str:
        # Ensure scheme so urlparse works
        if "://" not in url:
            url = "http://" + url
        extracted = tldextract.extract(url)
        return f"{extracted.domain}.{extracted.suffix}".lower()

    return get_base_domain(url1) == get_base_domain(url2)

class SecretResolver:
    """
    A resolver that always returns a secret value.
    Conforms to `Resolver` but you shouldn't call `.resolve()` on it.
    """
    def __init__(self, site_id: int | str | None):
        self.site_id = site_id
        self.site_secrets = SiteCredentialManager().load_site_with_secrets(site_id)

    def resolve(self, inp: ResolveInput) -> ResolveOutput:
        logger.debug(f"SecretResolver called with input: {inp.text} and context: {inp.context}")
        current_url = inp.context.get("current_url", None) if inp.context else None
        site_url = self.site_secrets.get("URL")
        logger.debug(f"Matching base domain and suffix of current URL {current_url} with site URL {site_url}")
                
        if current_url and site_url and base_domains_match(current_url, site_url):
            secret_value = self.site_secrets.get(inp.text)
            if secret_value:
                return ResolveOutput(value=secret_value)
        return ResolveOutput(value=None)

class DefaultDelegationResolver:
    """
    A resolver that always delegates to the user.
    Conforms to `Resolver` but you shouldn't call `.resolve()` on it.
    """
    def resolve(self, inp: ResolveInput) -> ResolveOutput:  # type: ignore[override]
        raise RuntimeError("DefaultDelegationResolver.resolve() should not be called.")
    
    def delegate_to_user(
        self,
        resp: LLMResponse,
        *,
        mutate: bool = False,
    ) -> LLMResponse:
        out = resp if mutate else deepcopy(resp)
        out.pop("step", None)
        out.pop("steps", None)
        out["step_execution"] = "DELEGATE_TO_USER"
        out["reasoning"] = "Delegated to user for input: " + out.get("reasoning", "No reasoning provided.")
        return out

# ---------- Minimal examples (optional) ----------
class EchoResolver:
    def resolve(self, inp: ResolveInput) -> ResolveOutput:
        logger.debug(f"EchoResolver called with input: {inp.text} and context: {inp.context}")
        return ResolveOutput(value=inp.text)

class NoOpResolver:
    def resolve(self, inp: ResolveInput) -> ResolveOutput:
        logger.debug(f"NoOpResolver called with input: {inp.text} and context: {inp.context}")
        return ResolveOutput(value=None)

def callable_with_context(prompt: str, ctx: Optional[Mapping[str, Any]] = None) -> str:
    logger.debug(f"callable_with_context called with prompt: {prompt}, context: {ctx}")
    if any(term in prompt.strip().lower() for term in ["source", "origin", "departure"]):
        return "New York"
    elif any(term in prompt.strip().lower() for term in ["destination", "target"]):
        return "Miami"
    return (ctx or {}).get(prompt, "Not Matched")



# surfari/mcp_client/manager.py
import asyncio
from typing import Dict, Optional, List, Any, Callable, Union

from surfari.model.mcp.types import MCPServerInfo, MCPTool, MCPResource, MCPCallResult
from surfari.model.mcp.session import MCPHTTPClientSession, MCPStdioFastMCPClientSession

# Unified session type (both share the same public API)
MCPAnySession = Union[MCPHTTPClientSession, MCPStdioFastMCPClientSession]


class MCPClientManager:
    def __init__(self, progress_cb: Optional[Callable[[str, int, int, str], None]] = None):
        # progress_cb kept for API compatibility (not used by current sessions)
        self._sessions: Dict[str, MCPAnySession] = {}
        self._progress_cb = progress_cb

    async def add_server(self, info: MCPServerInfo) -> None:
        """
        Add a server based on MCPServerInfo:
          - If `info.url` is present -> HTTP/SSE using MCPHTTPClientSession
          - Else -> STDIO using MCPStdioFastMCPClientSession with command/args
        """
        url = getattr(info, "url", None)
        if url:
            sess: MCPAnySession = MCPHTTPClientSession(url)
        else:
            sess = MCPStdioFastMCPClientSession(
                command=info.command,
                args=info.args or [],
                cwd=info.cwd or None,
                env=info.env or None,
            )

        await sess.connect()
        self._sessions[info.id] = sess

    def has_server(self, server_id: str) -> bool:
        return server_id in self._sessions

    async def list_tools(self, server_id: str) -> List[MCPTool]:
        return await self._sessions[server_id].list_tools()

    async def list_resources(self, server_id: str) -> List[MCPResource]:
        return await self._sessions[server_id].list_resources()

    async def read_resource(self, server_id: str, uri: str) -> MCPCallResult:
        return await self._sessions[server_id].read_resource(uri)

    async def call_tool(
        self,
        server_id: str,
        name: str,
        arguments: Dict[str, Any] | None = None,
        timeout_s: float | None = None,
    ) -> MCPCallResult:
        return await self._sessions[server_id].call_tool(name, arguments, timeout_s)

    async def aclose(self):
        await asyncio.gather(*(s.aclose() for s in self._sessions.values()), return_exceptions=True)
        self._sessions.clear()



from typing import Any, Dict, List, Optional, Tuple, Callable
import asyncio
import concurrent.futures
import json
import jsonschema
import time

from surfari.model.mcp.manager import MCPClientManager
from surfari.model.mcp.types import MCPTool, MCPCallResult
from surfari.util import surfari_logger as _surfari_logger

logger = _surfari_logger.getLogger(__name__)

SAFE_NAME = str.maketrans({c: "_" for c in " /:\\|@#?&%$!^*()[]{}<>,=+~`\""})
_EXECUTOR = concurrent.futures.ThreadPoolExecutor(max_workers=4)

def _run_sync(coro):
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return asyncio.run(coro)
    else:
        return _EXECUTOR.submit(asyncio.run, coro).result()

def _jsonschema_to_gemini(schema: Dict[str, Any]) -> Dict[str, Any]:
    if not schema:
        return {"type": "OBJECT"}
    t = (schema.get("type") or "object").lower()
    if t == "object":
        props = {}
        for k, v in (schema.get("properties") or {}).items():
            props[k] = _jsonschema_to_gemini(v)
            if "description" in v:
                props[k]["description"] = v["description"]
        out = {"type": "OBJECT", "properties": props}
        if "required" in schema:
            out["required"] = schema["required"]
        return out
    if t == "array":
        return {"type": "ARRAY", "items": _jsonschema_to_gemini(schema.get("items") or {"type": "string"})}
    if t in ("string",):
        return {"type": "STRING"}
    if t in ("integer",):
        return {"type": "INTEGER"}
    if t in ("number",):
        return {"type": "NUMBER"}
    if t in ("boolean",):
        return {"type": "BOOLEAN"}
    return {"type": "STRING"}

def _fn_name(server_id: str, tool_name: str) -> str:
    return f"mcp__{server_id.translate(SAFE_NAME)}__{tool_name.translate(SAFE_NAME)}"

def _reverse_name(fn_name: str) -> Tuple[str, str]:
    if not fn_name.startswith("mcp__"):
        raise ValueError("Not an MCP tool fn name")
    _, rest = fn_name.split("mcp__", 1)
    server_id, tool_name = rest.split("__", 1)
    return server_id, tool_name

class MCPToolRegistry:
    def __init__(self, manager: MCPClientManager):
        self.manager = manager
        self._by_fn: Dict[str, Tuple[str, MCPTool]] = {}

    async def refresh(self, server_ids: Optional[List[str]] = None) -> None:
        self._by_fn.clear()
        targets = server_ids or list(self.manager._sessions.keys())
        for sid in targets:
            tools = await self.manager.list_tools(sid)
            for t in tools:
                fn = _fn_name(sid, t.name)
                self._by_fn[fn] = (sid, t)

    def as_openai_tools(self) -> List[Dict[str, Any]]:
        out = []
        for fn, (_, tool) in self._by_fn.items():
            schema = tool.input_schema or {"type": "object", "properties": {}, "additionalProperties": True}
            out.append({
                "type": "function",
                "function": {
                    "name": fn[:64],
                    "description": (tool.description or "")[:512],
                    "parameters": schema
                }
            })
        return out

    def as_anthropic_tools(self) -> List[Dict[str, Any]]:
        out = []
        for fn, (_, tool) in self._by_fn.items():
            schema = tool.input_schema or {"type": "object", "properties": {}, "additionalProperties": True}
            out.append({
                "name": fn,
                "description": tool.description or "",
                "input_schema": schema
            })
        return out

    def as_gemini_function_declarations(self) -> List[Dict[str, Any]]:
        decls = []
        for fn, (_, tool) in self._by_fn.items():
            schema = tool.input_schema or {"type": "object", "properties": {}, "additionalProperties": True}
            decls.append({
                "name": fn[:64],
                "description": (tool.description or "")[:512],
                "parameters": _jsonschema_to_gemini(schema)
            })
        return decls

    def as_python_proxy_tools(self) -> list[Callable[..., Any]]:
        funcs: list[Callable[..., Any]] = []

        for fn_name, (_, mcp_tool) in self._by_fn.items():
            schema = mcp_tool.input_schema or {"type": "object", "properties": {}, "additionalProperties": True}

            def make_wrapper(bound_name: str, bound_schema: dict, bound_tool: MCPTool):
                def _mcp_proxy(**kwargs):
                    # optional per-call override: pass _timeout_s in kwargs if needed
                    timeout = kwargs.pop("_timeout_s", None)
                    try:
                        if bound_schema:
                            jsonschema.validate(instance=kwargs, schema=bound_schema)
                    except Exception as e:
                        return {"ok": False, "error": f"Schema validation failed: {e}"}
                    t0 = time.perf_counter()
                    res: MCPCallResult = _run_sync(self.execute(bound_name, kwargs, timeout_s=timeout))
                    dt = (time.perf_counter() - t0) * 1000
                    logger.debug("MCP proxy %s finished in %.1f ms (ok=%s)", bound_name, dt, getattr(res, "ok", None))                    
                    if res.ok:
                        try:
                            json.dumps(res.data)
                            return res.data
                        except Exception:
                            return json.loads(json.dumps(res.data, default=lambda o: repr(o)))
                    else:
                        return {"ok": False, "error": res.error}

                setattr(_mcp_proxy, "tool_name", bound_name)
                _mcp_proxy.__doc__ = (bound_tool.description or f"MCP tool '{bound_name}'")[:512]
                _mcp_proxy.__annotations__ = {k: Any for k in (bound_schema.get("properties") or {}).keys()}
                _mcp_proxy.__name__ = bound_name[:64]
                _mcp_proxy.__parameters_schema__ = bound_schema  # <— key line
                _mcp_proxy.__mcp_tool__ = bound_tool
                return _mcp_proxy

            funcs.append(make_wrapper(fn_name, schema, mcp_tool))
        return funcs


    async def execute(self, fn_name: str, arguments: Dict[str, Any] | None = None, timeout_s: Optional[float] = None) -> MCPCallResult:
        if fn_name not in self._by_fn:
            candidates = [k for k in self._by_fn if k.startswith(fn_name)]
            if len(candidates) == 1:
                fn_name = candidates[0]
            else:
                return MCPCallResult(ok=False, error=f"Unknown MCP tool: {fn_name}")

        server_id, tool = self._by_fn[fn_name]
        args = arguments or {}
        if tool.input_schema:
            try:
                jsonschema.validate(instance=args, schema=tool.input_schema)
            except Exception as e:
                return MCPCallResult(ok=False, error=f"Schema validation failed: {e}")

        return await self.manager.call_tool(server_id, tool.name, args, timeout_s)

    def has(self, fn_name: str) -> bool:
        return fn_name in self._by_fn

    def list_function_names(self) -> List[str]:
        return list(self._by_fn.keys())



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

@dataclass
class MCPTool:
    name: str
    description: str | None = None
    input_schema: Dict[str, Any] | None = None

@dataclass
class MCPResource:
    uri: str
    name: str
    description: str | None = None
    mime_type: str | None = None

@dataclass
class MCPServerInfo:
    id: str
    command: str
    args: List[str] = field(default_factory=list)
    env: Dict[str, str] = field(default_factory=dict)
    cwd: Optional[str] = None

@dataclass
class MCPCallResult:
    ok: bool
    data: Any = None
    error: Optional[str] = None
    elapsed_ms: Optional[int] = None



import re
import random
from typing import Set, Optional, Dict

import surfari.util.surfari_logger as surfari_logger
logger = surfari_logger.getLogger(__name__)

class NumericMasker:
    """
    Enhanced version that:
    - Never masks configurable numbers/terms (e.g., 1099, 2024, W2)
    - Maintains all existing date/time detection logic
    - Makes exclusion list and min length configurable
    - Optimizes unmasking by building the reverse map only once.
    """
    # Splits on whitespace only
    TOKEN_PATTERN = re.compile(r"\S+")

    # Potential month names
    MONTH_NAMES = (
        "Jan", "Feb", "Mar", "Apr", "May", "Jun",
        "Jul", "Aug", "Sep", "Sept", "Oct", "Nov", "Dec",
        "January", "February", "March", "April", "June",
        "July", "August", "September", "October",
        "November", "December"
    )
    MONTH_PATTERN = "|".join(MONTH_NAMES)

    # Time pattern: HH:MM(:SS)? plus optional AM/PM
    TIME_PATTERN = re.compile(r"^\d{1,2}:\d{2}(?::\d{2})?(?:\s?[APMapm]{2})?$")

    # Default donot mask terms (case-insensitive)
    DEFAULT_DONOT_MASK_TERMS = {"1099", "2024", "2025", "2026", "401k"}

    def __init__(self, donot_mask_items: Optional[Set[str]] = None, min_token_length: int = 5):
        """
        Initialize with optional set of donot mask terms and min token length.
        """
        self.donot_mask_terms = {
            re.sub(r"[^\w]", "", term).lower() for term in (donot_mask_items if donot_mask_items is not None else self.DEFAULT_DONOT_MASK_TERMS)
        }
        self.replacement_map: Dict[str, str] = {}  # old_token -> masked_token
        self.reverse_map: Dict[str, str] = {}      # masked_token -> old_token
        self.normalized_reverse_map: Dict[str, str] = {}  # normalized_masked_token -> normalized_old_token
        self.used_masked: Set[str] = set()         # set of masked tokens to avoid collisions
        self.min_token_length: int = min_token_length

    def _clear(self):
        self.replacement_map.clear()
        self.used_masked.clear()
        self.reverse_map.clear()
        self.normalized_reverse_map.clear()
        
    def _is_donot_mask_term(self, token: str) -> bool:
        """
        Check if token matches any donot mask term (case-insensitive),
        after normalizing the token (removing punctuation/special characters).
        """
        normalized = re.sub(r"[^\w]", "", token).lower()
        return normalized in self.donot_mask_terms

    def add_donot_mask_terms_from_string(self, in_string: str):
        """
        Tokenize the in_string and add digit-containing tokens meeting min length
        to donot_mask_terms (after normalization and lowercasing).
        """
        logger.debug(f"Adding donot mask terms from string: {in_string}")
        tokens = self.TOKEN_PATTERN.findall(in_string)
        for token in tokens:
            if self._has_digit(token) and len(token.strip()) >= self.min_token_length:
                normalized = re.sub(r"[^\w]", "", token).lower()
                if normalized:
                    self.donot_mask_terms.add(normalized)

    def _is_dateish(self, token: str) -> bool:
        t = token.strip()
        t2 = t[1:].lstrip() if t.startswith('-') else t

        if self.TIME_PATTERN.match(t2):
            return True
        if re.search(r"[/-]", t2) and re.search(r"\d", t2):
            return True
        if re.compile(rf"(?i)\b(?:{self.MONTH_PATTERN})\b").search(t2) and re.search(r"\d", t2):
            return True
        return False

    def _has_digit(self, token: str) -> bool:
        return any(ch.isdigit() for ch in token)

    def _mask_digit_char(self, c: str) -> str:
        """
        '0' -> random [0-9], '1'..'9' -> random [1-9]
        else unchanged
        """
        if c.isdigit():
            if c == '0':
                return random.choice("0123456789")
            else:
                return random.choice("123456789")
        return c

    def _mask_token(self, token: str) -> str:
        """If we've masked this token before, reuse it. Else create new masked token."""
        if token in self.replacement_map:
            return self.replacement_map[token]

        attempt_count = 0
        while True:
            attempt_count += 1
            if attempt_count > 100:
                fallback = f"MASK{random.randint(100000,999999)}"
                logger.warning(f"Exceeded attempts for token={token}, using {fallback}")
                self.replacement_map[token] = fallback
                self.used_masked.add(fallback)
                return fallback

            candidate = "".join(self._mask_digit_char(c) for c in token)
            if candidate not in self.used_masked:
                self.replacement_map[token] = candidate
                self.used_masked.add(candidate)
                return candidate
    
    def _strip_brackets(self, text: str) -> str:
        if text.startswith(("[]", "()", "[X]", "(X)")):
            return text
        
        while text and text[0] in "{[(":
            text = text[1:]
        while text and text[-1] in "}])":
            text = text[:-1]
        return text
    
    def _normalize_number(self, num_str: str) -> str:
        """
        Normalize a numeric string for comparison, preserving + or - if present.
        Removes commas and dollar signs, then float-casts and formats for comparison.
        """
        has_dollar = "$" in num_str
        prefix = ""

        num_str = re.sub(r"[{}\[\]\(\):;,$']", "", num_str).strip().rstrip(".")
        tmp = num_str
        if tmp.startswith(("+", "-")):
            prefix = tmp[0]
            tmp = tmp[1:].lstrip()

        try:
            val = float(tmp)
        except ValueError:
            return num_str  # Not a valid number, return semi-normalized

        # Format numbers as float with two decimal places or int
        formatted = f"{val:.2f}" if "." in tmp else f"{int(val)}"
        # add back $ and sign
        return f"{prefix}${formatted}" if has_dollar else f"{prefix}{formatted}"

    def _build_reverse_map(self):
        """Builds the reverse map from the replacement map for efficient lookup."""
        logger.debug("Building reverse map for sensitive info unmasking")
        self.reverse_map = {}
        for original, masked in self.replacement_map.items():
            self.reverse_map[masked] = original
            norm_masked = self._normalize_number(masked)
            norm_original = self._normalize_number(original)
            self.normalized_reverse_map[norm_masked] = norm_original
            logger.sensitive(f"added reverse mapping {masked} -> {original} and normalized reverse mapping: {norm_masked} -> {norm_original}")
                
    def mask_sensitive_info(self, text: str, donot_mask=[]) -> str:
        """
        Enhanced version that:
        1. Skips donot mask terms
        2. Skips dateish tokens
        3. Masks other digit-containing tokens
        """     
        def replacer(m):
            token = m.group(0)
            norm_token = self._normalize_number(token)
            if len(norm_token) < self.min_token_length:
                return token
            if not self._has_digit(token):
                return token
            if self._is_donot_mask_term(token):
                return token
            if self._is_dateish(token):
                return token
            for item in donot_mask:
                if token in item:
                    return token
            return self._mask_token(token)

        logger.info(f"Masking sensitive info called for text of {len(text)} characters")
        self._clear()  # Clear previous state
        masked_text = self.TOKEN_PATTERN.sub(replacer, text)
        self._build_reverse_map()  # Build reverse map after masking is complete
        logger.debug(f"Reverse map size: {len(self.reverse_map)}")
        return masked_text

    def unmask_sensitive_info(self, masked_text: str):
        """
        Revert masked tokens if they appear in self.reverse_map.
        """
        if not masked_text:
            return ""
        
        logger.debug(f"unmasking sensitive info from: {masked_text}")
        
        def revert_func(m):
            token = m.group(0)
            if not self._has_digit(token) or self._is_dateish(token):
                return token
            original = self.reverse_map.get(token)
            if original:
                logger.sensitive(f"Revert token to original directly '{token}' -> '{original}'")
                return original
            norm_token = self._normalize_number(token)
            norm_original = self.normalized_reverse_map.get(norm_token)
            if norm_original:
                logger.sensitive(f"Revert token after normalizing '{token}' to {norm_token} -> '{norm_original}'")
                return norm_original
            logger.debug(f"Revert token FAILED to find original for '{token}'")
            return token

        unmasked_text = self.TOKEN_PATTERN.sub(revert_func, masked_text)
        logger.sensitive(f"unmasked text: {unmasked_text}")
        if not re.fullmatch(r"^(\[\[.*\]\]|\{\{.*\}\}|\[.*\]|\{.*\})\d*$", masked_text):
            # masked_text itself is not a proper target
            return re.sub(r"[{}\[\]\(\)]", "", unmasked_text)
        return unmasked_text



import os
import json
import base64
import asyncio
import aiofiles
import httpx
from email.message import EmailMessage
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Dict, Any, List
from urllib.parse import quote

from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request

from surfari.util.config import PROJECT_ROOT
import surfari.util.surfari_logger as surfari_logger
logger = surfari_logger.getLogger(__name__)


# --- Sheets scopes (reused via the generic scope upgrader) ---
SHEETS_SCOPE_READONLY = "https://www.googleapis.com/auth/spreadsheets.readonly"
SHEETS_SCOPE_RW = "https://www.googleapis.com/auth/spreadsheets"


class GmailClientAsync:
    """
    Minimal async Gmail client supporting send + read via Gmail REST API,
    plus lightweight Google Sheets helpers.

    - Tokens stored at:   PROJECT_ROOT/security/google_auth_token.json
    - Client secrets at:  PROJECT_ROOT/security/google_client_secret.json

    This version supports:
      * Automatic scope union/upgrade when an API requires broader permissions
      * Automatic single-retry on 401 (expired/invalid) and 403 insufficientPermissions
    """
    SCOPES = [
        "https://www.googleapis.com/auth/gmail.readonly",
        "https://www.googleapis.com/auth/gmail.send",
    ]

    def __init__(
        self,
        token_file: str = "google_auth_token.json",
        secrets_file: str = "google_client_secret.json",
        *,
        prefer_console: Optional[bool] = None,
    ):
        self.token_file = os.path.join(PROJECT_ROOT, "security", token_file)
        self.secrets_file = os.path.join(PROJECT_ROOT, "security", secrets_file)
        self.creds: Optional[Credentials] = None
        self.executor = ThreadPoolExecutor(max_workers=2)

        # Serialize all auth/refresh so only one flow occurs at a time
        self._auth_lock = asyncio.Lock()

        # Console-flow toggle (env wins if not explicitly provided)
        if prefer_console is None:
            env_val = os.getenv("GMAIL_OAUTH_CONSOLE", "0").lower()
            self.prefer_console = env_val in ("1", "true", "yes", "on")
        else:
            self.prefer_console = bool(prefer_console)

    async def _refresh_creds_async(self) -> None:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(self.executor, self.creds.refresh, Request())

    async def _run_oauth_flow(self, scopes: Optional[List[str]] = None) -> Credentials:
        """
        Launches the installed-app OAuth flow. Prefer an ephemeral local server,
        fall back to console flow if the port is unavailable or console is requested.
        Runs in a worker thread to avoid blocking the event loop.
        """
        scopes = scopes or self.SCOPES
        logger.debug("[🔐] Starting OAuth flow for scopes: %s", scopes)
        flow = InstalledAppFlow.from_client_secrets_file(self.secrets_file, scopes)

        loop = asyncio.get_running_loop()

        # If explicitly preferring console, go straight there.
        if self.prefer_console:
            logger.debug("[🔐] Using console OAuth flow (GMAIL_OAUTH_CONSOLE=1 or prefer_console=True)")
            return await loop.run_in_executor(self.executor, flow.run_console)

        def _local_server():
            # Ephemeral port (0) + loopback host; open browser if possible.
            return flow.run_local_server(host="127.0.0.1", port=0, open_browser=True, timeout_seconds=300)

        try:
            return await loop.run_in_executor(self.executor, _local_server)
        except OSError as e:
            # Typical: OSError: [Errno 48] Address already in use, or environment without browser
            logger.warning("Local OAuth server failed (%s). Falling back to console flow.", e)
            return await loop.run_in_executor(self.executor, flow.run_console)

    async def _ensure_scopes(self, desired_scopes: List[str]) -> None:
        """
        Ensure credentials include desired_scopes.
        If scopes are missing, re-run OAuth with the union of existing+desired scopes.
        Also handles refresh when token is expired.
        """
        desired = sorted(set(desired_scopes))
        async with self._auth_lock:
            existing_scopes: List[str] = []

            # Load existing token if present
            if os.path.exists(self.token_file):
                async with aiofiles.open(self.token_file, "r") as f:
                    token_data = await f.read()
                    try:
                        self.creds = Credentials.from_authorized_user_info(json.loads(token_data))
                        existing_scopes = list(self.creds.scopes or [])
                    except Exception as e:
                        logger.warning("Failed to load existing google_auth_token.json; will re-consent. Error: %s", e)
                        self.creds = None

            have = set(existing_scopes)
            need_upgrade = not set(desired).issubset(have)

            # If no creds OR invalid, try refresh (when appropriate) or re-consent
            if not self.creds or not self.creds.valid or need_upgrade:
                all_scopes = sorted(set(existing_scopes) | set(desired))
                if self.creds and self.creds.expired and self.creds.refresh_token and not need_upgrade:
                    logger.debug("[🔄] Refreshing access token…")
                    try:
                        await self._refresh_creds_async()
                    except Exception as e:
                        logger.error("Token refresh failed: %s", e)
                        self.creds = await self._run_oauth_flow(all_scopes)
                else:
                    logger.info("[🔐] Running OAuth flow for scopes: %s", all_scopes)
                    self.creds = await self._run_oauth_flow(all_scopes)

                # Persist updated credentials
                async with aiofiles.open(self.token_file, "w") as f:
                    await f.write(self.creds.to_json())

    async def authenticate(self) -> None:
        # Keep existing call sites working: ensure base Gmail scopes.
        await self._ensure_scopes(self.SCOPES)

    async def _request_json(
        self,
        method: str,
        url: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        json_body: Optional[Dict[str, Any]] = None,
        need_scopes: Optional[List[str]] = None,
        timeout: float = 30.0,
        retry_on_401_403: bool = True,
    ) -> Dict[str, Any]:
        """
        Make an HTTP request to Google APIs with current creds, parse JSON, and
        auto-recover once from 401 invalid/expired token and 403 insufficientPermissions.

        - need_scopes: scopes that this endpoint requires (used to auto-upgrade on 403).
        - Returns: {"ok": True, "json": <parsed_json>} on success OR {"ok": False, "status": <int>, "error": <str>} on error.
        """
        # Ensure at least base Gmail scopes so self.creds is set (and token refreshed if needed).
        # If need_scopes is provided and goes beyond base, _ensure_scopes will union-upgrade as necessary.
        scopes = list(set((need_scopes or []) + self.SCOPES))
        await self._ensure_scopes(scopes)

        headers = {"Authorization": f"Bearer {self.creds.token}"}
        attempt = 0

        async with httpx.AsyncClient(timeout=timeout) as client:
            while True:
                attempt += 1
                try:
                    r = await client.request(method.upper(), url, headers=headers, params=params, json=json_body)
                    r.raise_for_status()
                    try:
                        return {"ok": True, "json": r.json()}
                    except Exception:
                        return {"ok": True, "json": {}}

                except httpx.HTTPStatusError as e:
                    status = e.response.status_code
                    text = e.response.text[:500]

                    # Try to parse Google error to detect reason
                    reason = ""
                    try:
                        ej = e.response.json()
                        reason = (ej.get("error", {}) or {}).get("status") or (ej.get("error", {}) or {}).get("message", "")
                        if not reason:
                            errs = (ej.get("error", {}) or {}).get("errors", [])
                            if errs and isinstance(errs, list):
                                reason = errs[0].get("reason", "") or errs[0].get("message", "")
                    except Exception:
                        pass

                    can_retry = retry_on_401_403 and attempt == 1  # retry only once
                    if can_retry and status == 401:
                        # Common: invalid/expired token; try refresh or full flow with current scopes
                        logger.info("[🔁] 401 received; attempting token refresh and retry…")
                        try:
                            await self._refresh_creds_async()
                            headers["Authorization"] = f"Bearer {self.creds.token}"
                            continue
                        except Exception as e2:
                            logger.warning("Refresh failed; re-consenting with current scopes. Error: %s", e2)
                            await self._ensure_scopes(scopes)
                            headers["Authorization"] = f"Bearer {self.creds.token}"
                            continue

                    if can_retry and status == 403 and "insufficient" in (reason or "").lower():
                        # We need to upgrade scopes for this operation
                        upgrade_scopes = need_scopes or []
                        if upgrade_scopes:
                            logger.info("[🔁] 403 insufficientPermissions; upgrading scopes: %s", upgrade_scopes)
                            await self._ensure_scopes(list(set(self.SCOPES + upgrade_scopes)))
                            headers["Authorization"] = f"Bearer {self.creds.token}"
                            continue

                    # If we reach here, no more retries
                    return {"ok": False, "status": status, "error": text}

                except Exception as e:
                    # Network/parse etc. Non-retryable here.
                    return {"ok": False, "status": -1, "error": str(e)}

    # ---------------------- READ ----------------------

    async def search_emails(self, query: str, max_results: int = 5) -> Dict[str, Any]:
        """
        Search messages by Gmail query. Returns lightweight info:
          {"ok": True, "messages": [{"id","threadId","snippet","headers":{…}}]}
        """
        await self.authenticate()
        list_url = "https://gmail.googleapis.com/gmail/v1/users/me/messages"
        list_params = {"q": query, "maxResults": max(1, int(max_results))}
        logger.debug("[>] Gmail search: %s", list_params)

        r1 = await self._request_json("GET", list_url, params=list_params, timeout=30.0)
        if not r1.get("ok"):
            return {"ok": False, "error": r1.get("error", ""), "status": r1.get("status")}
        items = r1["json"].get("messages", []) or []
        out: List[Dict[str, Any]] = []

        # Fetch metadata for each message
        for m in items:
            mid = m.get("id")
            if not mid:
                continue
            msg_url = f"https://gmail.googleapis.com/gmail/v1/users/me/messages/{mid}"
            params = {"format": "metadata", "metadataHeaders": ["From", "To", "Subject", "Date"]}
            r2 = await self._request_json("GET", msg_url, params=params, timeout=30.0)
            if not r2.get("ok"):
                out.append({"id": mid, "error": r2.get("error", ""), "status": r2.get("status")})
                continue
            j = r2["json"]
            hdrs = {h["name"]: h["value"] for h in j.get("payload", {}).get("headers", [])}
            out.append({
                "id": j.get("id"),
                "threadId": j.get("threadId"),
                "snippet": j.get("snippet"),
                "headers": {
                    "From": hdrs.get("From"),
                    "To": hdrs.get("To"),
                    "Subject": hdrs.get("Subject"),
                    "Date": hdrs.get("Date"),
                },
            })

        return {"ok": True, "messages": out}

    async def get_message(self, message_id: str) -> Dict[str, Any]:
        """
        Get a single message (metadata + snippet). Returns {"ok": True, "message": {...}}.
        """
        await self.authenticate()
        url = f"https://gmail.googleapis.com/gmail/v1/users/me/messages/{message_id}"
        params = {"format": "metadata", "metadataHeaders": ["From", "To", "Subject", "Date"]}

        r = await self._request_json("GET", url, params=params, timeout=30.0)
        if not r.get("ok"):
            return {"ok": False, "error": r.get("error", ""), "status": r.get("status")}
        j = r["json"]
        hdrs = {h["name"]: h["value"] for h in j.get("payload", {}).get("headers", [])}
        out = {
            "id": j.get("id"),
            "threadId": j.get("threadId"),
            "snippet": j.get("snippet"),
            "headers": {
                "From": hdrs.get("From"),
                "To": hdrs.get("To"),
                "Subject": hdrs.get("Subject"),
                "Date": hdrs.get("Date"),
            },
        }
        return {"ok": True, "message": out}

    # ---------------------- SEND ----------------------

    async def send_email(
        self,
        to: str,
        subject: str,
        body: str,
        *,
        cc: Optional[str] = None,
        bcc: Optional[str] = None,
        html: bool = False,
        from_addr: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Send an email via Gmail. Returns {"ok": True, "id": "<gmail_message_id>"} on success.
        """
        await self.authenticate()
        msg = EmailMessage()
        if from_addr:
            msg["From"] = from_addr
        msg["To"] = to
        if cc:
            msg["Cc"] = cc
        if bcc:
            msg["Bcc"] = bcc
        msg["Subject"] = subject

        if html:
            msg.add_alternative(body, subtype="html")
        else:
            msg.set_content(body)

        raw = base64.urlsafe_b64encode(msg.as_bytes()).decode("utf-8").rstrip("=")

        url = "https://gmail.googleapis.com/gmail/v1/users/me/messages/send"
        payload = {"raw": raw}

        r = await self._request_json("POST", url, json_body=payload, timeout=30.0)
        if not r.get("ok"):
            return {"ok": False, "error": r.get("error", ""), "status": r.get("status")}
        return {"ok": True, "id": r["json"].get("id")}

    # ---------------------- SHEETS: READ → JSON ----------------------

    async def sheets_read_to_json(
        self,
        spreadsheet_id: str,
        range_a1: str,
        header_row: int = 1,
    ) -> Dict[str, Any]:
        """
        Read a Sheets range and convert tabular data to list-of-dicts.
        Assumes the first row of the fetched range is the header.
        """
        need_scopes = [SHEETS_SCOPE_READONLY]

        url = f"https://sheets.googleapis.com/v4/spreadsheets/{spreadsheet_id}/values/{quote(range_a1, safe='!$:,A-Za-z0-9')}"
        params = {"majorDimension": "ROWS"}

        r = await self._request_json("GET", url, params=params, need_scopes=need_scopes, timeout=30.0)
        if not r.get("ok"):
            return {"ok": False, "error": r.get("error", ""), "status": r.get("status")}

        payload = r["json"]
        values = payload.get("values", []) or []
        if not values:
            return {"ok": True, "rows": [], "columns": [], "count": 0}

        # Take the first row in the returned values as header
        columns = [str(c).strip() for c in (values[0] if values else [])]
        data_rows = values[1:] if len(values) > 1 else []

        # Build list-of-dicts; pad/truncate rows to header length
        rows: List[Dict[str, Any]] = []
        for raw in data_rows:
            row = list(raw)
            if len(row) < len(columns):
                row += [""] * (len(columns) - len(row))
            elif len(row) > len(columns):
                row = row[: len(columns)]
            rows.append({col: cell for col, cell in zip(columns, row)})

        return {"ok": True, "rows": rows, "columns": columns, "count": len(rows)}

    # ---------------------- SHEETS: CREATE from JSON ----------------------

    async def sheets_create_from_json(
        self,
        title: str,
        records: List[Dict[str, Any]],
        sheet_name: str = "Sheet1",
    ) -> Dict[str, Any]:
        """
        Create a new Google Sheet and populate it from a list of dicts.
        - 'title' creates the spreadsheet with that name.
        - 'records' is a list of dicts; columns are the union of keys (first-seen order).
        - 'sheet_name' sets the first sheet's title (default: Sheet1).
        """
        need_scopes = [SHEETS_SCOPE_RW]

        # 1) Create spreadsheet
        create_url = "https://sheets.googleapis.com/v4/spreadsheets"
        body_create = {"properties": {"title": title}, "sheets": [{"properties": {"title": sheet_name}}]}

        cr = await self._request_json("POST", create_url, json_body=body_create, need_scopes=need_scopes, timeout=60.0)
        if not cr.get("ok"):
            return {"ok": False, "error": cr.get("error", ""), "status": cr.get("status")}

        created = cr["json"]
        spreadsheet_id = created.get("spreadsheetId")
        if not spreadsheet_id:
            return {"ok": False, "error": "Create returned no spreadsheetId"}

        # 2) Prepare values (header + rows)
        columns: List[str] = []
        for rec in (records or []):
            for k in rec.keys():
                if k not in columns:
                    columns.append(k)

        values: List[List[Any]] = [columns]
        for rec in (records or []):
            values.append([rec.get(c, "") for c in columns])

        # 3) Write values via values.update
        rng = f"{sheet_name}!A1"
        update_url = f"https://sheets.googleapis.com/v4/spreadsheets/{spreadsheet_id}/values/{quote(rng, safe='!$:,A-Za-z0-9')}"
        params = {"valueInputOption": "USER_ENTERED"}  # respect user formatting
        body_update = {"range": rng, "majorDimension": "ROWS", "values": values}

        ur = await self._request_json("PUT", update_url, params=params, json_body=body_update, need_scopes=need_scopes, timeout=60.0)
        if not ur.get("ok"):
            return {"ok": False, "error": ur.get("error", ""), "status": ur.get("status")}

        return {
            "ok": True,
            "spreadsheetId": spreadsheet_id,
            "sheetTitle": sheet_name,
            "columns": columns,
            "count": len(records or []),
        }


# ---------------------- Tool callables ----------------------

_gmail_singleton: Optional[GmailClientAsync] = None

def _client() -> GmailClientAsync:
    global _gmail_singleton
    if _gmail_singleton is None:
        _gmail_singleton = GmailClientAsync()
    return _gmail_singleton

# Keep signatures simple (AFC/OpenAI friendly) and return JSON-serializable dicts.

async def gmail_send_email(to: str, subject: str, body: str, cc: str = "", bcc: str = "", html: bool = False) -> Dict[str, Any]:
    """
    Send an email using the authenticated Gmail account.
    - to: comma-separated recipients
    - subject: subject line
    - body: plain text (or HTML if html=True)
    - cc, bcc: optional comma-separated lists
    - html: set True to send HTML body
    """
    return await _client().send_email(to=to, subject=subject, body=body, cc=(cc or None), bcc=(bcc or None), html=bool(html))

async def gmail_search_emails(query: str, max_results: int = 5) -> Dict[str, Any]:
    """
    Search inbox with a Gmail query (e.g., 'from:me after:1723950000', 'subject:OTP').
    Returns compact message list.
    """
    return await _client().search_emails(query=query, max_results=max_results)

async def gmail_get_message(message_id: str) -> Dict[str, Any]:
    """
    Fetch a single message by ID (metadata + snippet).
    """
    return await _client().get_message(message_id=message_id)

# --- Sheets tools ---

async def sheets_read_json(spreadsheet_id: str, range_a1: str, header_row: int = 1) -> Dict[str, Any]:
    """
    Read a Google Sheets range (A1 notation) and return JSON rows.
    - spreadsheet_id: the Sheet's ID (from URL)
    - range_a1: e.g., 'Sheet1!A1:D200' (first row is treated as header)
    - header_row: kept for API parity; header is taken from the first returned row
    """
    return await _client().sheets_read_to_json(spreadsheet_id=spreadsheet_id, range_a1=range_a1, header_row=header_row)

async def sheets_create_from_json(title: str, records: List[Dict[str, Any]], sheet_name: str = "Sheet1") -> Dict[str, Any]:
    """
    Create a Google Sheet and populate it with records (list of dicts).
    - title: spreadsheet title
    - records: list of JSON dicts (keys become columns)
    - sheet_name: first sheet/tab name
    """
    return await _client().sheets_create_from_json(title=title, records=records, sheet_name=sheet_name)

TOOLS = [gmail_send_email, gmail_search_emails, gmail_get_message, sheets_read_json, sheets_create_from_json]

if __name__ == "__main__":
    from surfari.model.tool_helper import _normalize_tools_for_openai

    normalized = _normalize_tools_for_openai(TOOLS)
    import json
    print(json.dumps(normalized, indent=2))


import re
import surfari.util.config as config
import surfari.util.surfari_logger as surfari_logger
logger = surfari_logger.getLogger(__name__)

def rearrange_texts(input_str, y_threshold=16, h_scale_factor=4, additional_text=""):
    """
    Parses an input string containing lines of the form:
      [frame_name] <text> (x=<val>, y=<val>, w=<val>, h=<val>, xpath=<val>)
    or
      <text> (x=<val>, y=<val>, w=<val>, h=<val>, xpath=<val>, locator_string=<empty_or_not>)

    Groups lines with similar y-coordinates (within y_threshold) into rows,
    sorts each row by x-coordinate, and for each row builds an output line where:

      - Each text item is placed at a column proportional to its x coordinate,
        computed as: target_column = max(0, int(round(x / h_scale_factor))).

      - If multiple items share (y, x), we preserve original input order via _orig_index.

      - Only wraps text at word boundaries if the text length is more than a wrapping factor *
        the "max_col_width," computed from w/h_scale_factor.

      - Excessive internal whitespace is replaced with a single space.

      - Between rows, extra blank lines are inserted if the vertical gap
        exceeds y_threshold, pushing lines downward.
    """

    lines = input_str.strip().splitlines()

    # Regex captures optional frame name, text, x, y, w, h, and xpath
    pattern = re.compile(
        r"^(?:([^\s]+)\s+)?(.*?)\s*\(x=([-\d\.]+),\s*y=([-\d\.]+),\s*w=([\d\.]+),\s*h=([\d\.]+),\s*xpath=(.*?),\s*locator_string=(.*?)\)$"
    )

    entries = []
    for i, line in enumerate(lines):
        line = line.strip()
        if "(x=" not in line or "xpath=" not in line:
            logger.warning(f"Skipping line without expected coordinates/xpath: {line}")
            continue

        match = pattern.match(line)
        if match:
            frame_name = match.group(1) or ""
            raw_text = match.group(2).strip()
            # Collapse all whitespace runs into a single space
            text = re.sub(r"\s+", " ", raw_text)

            x = float(match.group(3))
            y = float(match.group(4))
            w = float(match.group(5))
            h = float(match.group(6))
            xpath = match.group(7).strip()
            
            # allow 1x1 radio, checkbox, or button
            if (w <=1 or h <= 1) and len(text) > 5:
                logger.debug(f"Skipping off-screen or zero-width text: {text}, x={x}, y={y}, w={w}, h={h}")
                continue # Skip off-screen elements          
             
            entries.append(
                {
                    "_orig_index": i,  # preserve input order as a tiebreaker
                    "frame_name": frame_name,
                    "text": text,
                    "x": x,
                    "y": y,
                    "w": w,
                    "h": h,
                    "xpath": xpath,
                }
            )
        else:
            logger.error(f"Line did not match expected pattern: {line}")

    month_pattern = re.compile(
        r"\b("
        r"January|February|March|April|May|June|July|August|September|October|November|December|"
        r"Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Sept|Oct|Nov|Dec"
        r")(\s?\d{4})?\b",
        re.IGNORECASE
    )
    # 1–2 digit tokens, optionally inside [ ]
    day_token_pattern = re.compile(r"(?:\[\s*)?\b\d{1,2}\b(?:\s*\])?")

    # Indices of entries that are pure month headers
    month_header_indices = [i for i, e in enumerate(entries) if month_pattern.fullmatch(e["text"])]

    month_offset_y = 0
    start_pair_idx = None  # index into month_header_indices (the first month of the validated pair)

    if len(month_header_indices) >= 2:
        # Find the first consecutive pair that has >= 5 day tokens between them
        for i in range(len(month_header_indices) - 1):
            s = month_header_indices[i]
            t = month_header_indices[i + 1]
            between = entries[s:t]  # entries in the first month block
            day_like_count = sum(len(day_token_pattern.findall(e["text"])) for e in between)
            if day_like_count >= 5:
                start_pair_idx = i
                # Compute block height from the validated block
                if between:
                    min_y = min(e["y"] for e in between)
                    max_y = max(e["y"] + e["h"] for e in between)
                else:
                    # Edge: no lines between headers; use headers' boxes
                    min_y = entries[s]["y"]
                    max_y = entries[t]["y"] + entries[t]["h"]
                month_offset_y = (max_y - min_y) + 40
                break

    if start_pair_idx is not None and month_offset_y > 0:
        # Build block boundaries from the validated first month onward:
        # [start_of_block_0 (=first month), start_of_block_1 (=second month), ..., end (=len(entries))]
        boundaries = month_header_indices[start_pair_idx:] + [len(entries)]

        # Shift each block after the first by N * month_offset_y
        # block 0 gets 0 * offset, block 1 gets 1 * offset, etc.
        for block_idx in range(1, len(boundaries) - 1):
            block_start = boundaries[block_idx]
            block_end   = boundaries[block_idx + 1]
            shift = month_offset_y * block_idx
            for j in range(block_start, block_end):
                entries[j]["y"] += shift
            
    # Sort by (y, x, _orig_index) so items with identical y/x come in file order
    entries.sort(key=lambda e: (e["y"], e["x"], e["_orig_index"]))

    # Group into rows based on the row's min y
    # (If an item is more than y_threshold above the row's min y, start a new row)
    rows: list[list[dict]] = []
    X_NEAR = 320          # ← tune if needed; smaller ⇒ stricter “same line”
    EPS    = 1e-3         # ← float-noise tolerance for “exactly the same y”

    for entry in entries:
        # rows whose baseline-y is within y_threshold of this entry
        candidate_rows = [
            (idx, row) for idx, row in enumerate(rows)
            if abs(entry["y"] - min(e["y"] for e in row)) < y_threshold
        ]

        if not candidate_rows:                # -- first element of a new row
            rows.append([entry])
            continue

        # ➊ if any candidate row has essentially the *same y*, use it directly
        for idx, row in candidate_rows:
            if abs(entry["y"] - min(e["y"] for e in row)) < EPS:
                rows[idx].append(entry)
                break
        else:
            # ➋ otherwise pick the row whose closest x-distance is minimal
            best_idx, best_dist = None, float("inf")
            for idx, row in candidate_rows:
                dist = min(abs(entry["x"] - e["x"]) for e in row)
                if dist < best_dist:
                    best_dist = dist
                    best_idx = idx

            if best_dist < X_NEAR:
                rows[best_idx].append(entry)
            else:
                rows.append([entry])

    # Sort each row by x & _orig_index again, then store row info
    row_data = []
    for row in rows:
        row.sort(key=lambda e: (e["x"], e["_orig_index"]))
        row_min_y = min(e["y"] for e in row)
        row_data.append({"entries": row, "row_min_y": row_min_y})

    output_lines = []
    prev_row_min_y = None
    prev_row_height = 0  # Track how many lines the previous row occupied

    for row in row_data:
        row_min_y = row["row_min_y"]
                    
        # Calculate vertical gap from previous content (not just previous row min y)
        if prev_row_min_y is not None:
            # The actual bottom of previous content is prev_row_min_y + (prev_row_height * y_threshold)
            content_bottom = prev_row_min_y + (prev_row_height * y_threshold)
            gap = row_min_y - content_bottom
            
            # Calculate how many blank lines to add based on multiples of y_threshold
            if gap >= y_threshold:
                multiple = gap / y_threshold
                n_blank = int(multiple)  # Add one blank line per multiple of y_threshold
                for _ in range(n_blank):
                    output_lines.append("")                    

        # Track each entry's wrapped lines
        entry_lines = []
        max_wrapped_lines = 1
        
        for e in row["entries"]:
            max_col_width = max(1, int(round(e["w"] / h_scale_factor)))
            target_col = max(0, int(round(e["x"] / h_scale_factor)))
            
            wrapping_factor = 1.0
            if len(e["text"]) <= 40:
                wrapping_factor = 1.8
            if len(e["text"]) <= 6: # case like [1], [AI], {{2}}, {{AI}}
                wrapping_factor = 6
                
            if "||" in e["text"] or len(e["text"]) > max_col_width * wrapping_factor:
                wrapped = word_wrap(e["text"], max_col_width)
            else:
                wrapped = [e["text"]]
            
            entry_lines.append({
                "lines": wrapped,
                "target_col": target_col
            })
            max_wrapped_lines = max(max_wrapped_lines, len(wrapped))

        # Build row lines
        row_lines = [""] * max_wrapped_lines
        for entry in entry_lines:
            for i, line in enumerate(entry["lines"]):
                if i < len(row_lines):
                    row_lines[i] = place_text(row_lines[i], line, entry["target_col"])
                else:
                    row_lines.append(place_text("", line, entry["target_col"]))
        
        output_lines.extend(row_lines)
        prev_row_min_y = row_min_y
        prev_row_height = max_wrapped_lines  # Update for next iteration

    output = "\n".join(output_lines)
    output = _remove_excessive_whitespace(output)
    if additional_text:
        output = additional_text + "\n" + output
    return output

def _remove_excessive_whitespace(text: str) -> str:
    """
    Remove excessive internal whitespace from the text.
    Remove more than 3 consecutive empty lines to reduce to 3.
    """
    return re.sub(r'((?:\r?\n){4,})', '\n\n\n', text)

def place_text(existing_line: str, text: str, target: int) -> str:
    """
    Place 'text' into 'existing_line' at column 'target'.
    If existing_line is shorter, we pad with spaces.
    If existing_line is already long, we add a space if needed,
    then append the text.
    """
    if len(existing_line) < target:
        existing_line += " " * (target - len(existing_line))
    else:
        if existing_line and not existing_line.endswith(" "):
            existing_line += " "
    return existing_line + text

def word_wrap(text: str, max_width: int) -> list:
    """
    Splits 'text' into multiple lines, never exceeding 'max_width' characters.
    Wraps at word boundaries. If '||' is found, force line breaks at those points.
    """
    if "||" in text:
        segments = [s.strip() for s in text.split("||")]   # keep empties for check
        wrapped_lines = []
        for seg in segments:
            seg = seg.strip()
            if seg == "-" or not seg:          # ① skip the lone dash (or blanks)
                continue
            wrapped_lines.extend(word_wrap(seg, max_width))   # recurse / normal wrap
        return wrapped_lines

    # Regular word wrapping
    words = text.split()
    lines = []
    current_line = ""

    for w in words:
        # if adding w + space to current_line would exceed max_width
        if current_line:
            if len(current_line) + 1 + len(w) <= max_width:
                current_line += " " + w
            else:
                lines.append(current_line)
                current_line = w
        else:
            # if w is bigger than max_width, forcibly break
            if len(w) > max_width:
                forced_lines = [
                    w[i : i + max_width] for i in range(0, len(w), max_width)
                ]
                if forced_lines:
                    current_line = forced_lines[0]
                    for chunk in forced_lines[1:]:
                        lines.append(current_line)
                        current_line = chunk
            else:
                current_line = w

    if current_line:
        lines.append(current_line)

    return lines



import re
import json
import asyncio
import os
import random
import time
from playwright.async_api import Locator
from typing import Dict, Tuple, Iterable

import surfari.util.config as config
import surfari.util.surfari_logger as surfari_logger
logger = surfari_logger.getLogger(__name__)

COUNT_WITHOUT_REASONING = """
(() => {
  const total = document.querySelectorAll('*').length;
  const overlays = document.querySelectorAll('#__surfari_reasoning_box__').length;
  return total - overlays;
})();
"""

def remove_unescaped_control_characters(json_str):
    """
    Remove unescaped control characters (ASCII 0-31) from a JSON string.

    This function uses a regex to find control characters (characters in the range
    U+0000 to U+001F) that are not already escaped (i.e. not immediately preceded by a backslash)
    and removes them.

    Note: This may result in loss of some whitespace formatting in string values.
    """
    # The regex (?<!\\)[\x00-\x1F] matches any control character not preceded by a backslash.
    pattern = re.compile(r"(?<!\\)[\x00-\x1F]")
    return pattern.sub("", json_str)

async def wait_for_page_load_generic(page, timeout_ms=1000, post_load_timeout_ms=2000):
    """
    Waits for page to load, stabilize DOM, and final JavaScript execution.

    Args:
        page: Playwright page object.
        timeout_ms: Max time to wait in milliseconds.
        post_load_timeout_ms: Final buffer after load.
    """
    import asyncio
    import time
    start_time = time.time()
    try:
        # 1. Wait for 'load' event
        await page.wait_for_load_state("load", timeout=timeout_ms)
        logger.debug("Page load state 'load' reached.")

        # 2. Wait for DOM stabilization
        dom_stable = await wait_for_dom_stable(page, timeout=timeout_ms)
        logger.debug("Page load state DOM structure stabilized.")
        
        # 3. Wait for network idle
        network_max_inflight = config.CONFIG["app"].get("network_max_inflight", 1)
        network_idle_quiet_ms = config.CONFIG["app"].get("network_idle_quiet_ms", 200)
        network_idle_timeout_ms = config.CONFIG["app"].get("network_idle_timeout_ms", 10000)
        
        if network_idle_timeout_ms > 0:
            logger.debug("Waiting for network to become quiet...")
            await wait_for_network_quiet(page, max_inflight=network_max_inflight, quiet_ms=network_idle_quiet_ms, timeout_ms=network_idle_timeout_ms)
            logger.debug("Page load state 'networkidle' reached.")
            
    except Exception as e:
        logger.error(f"Page load state failed: {e}")

    if (post_load_timeout_ms > 0):
        await asyncio.sleep(post_load_timeout_ms / 1000)
        logger.debug(f"Page load state compensation timeout of {post_load_timeout_ms}ms completed.")
        
    total_time = time.time() - start_time
    logger.debug(f"Page load state total complete after {total_time:.2f} seconds.")

async def wait_for_dom_stable(page, timeout=3000):
    logger.debug("Polling DOM element count manually...")

    start_time = time.time()
    end_time = start_time + timeout / 1000

    prev_count = None

    while time.time() < end_time:
        try:
            count = await page.evaluate(COUNT_WITHOUT_REASONING)
            if prev_count is not None and count == prev_count:
                logger.debug(f"DOM element count stabilized at {count}")
                return True
            prev_count = count
        except Exception as e:
            logger.warning(f"Error while polling DOM elements: {e}")
        
        await asyncio.sleep(0.2)  # Poll every 200ms

    raise TimeoutError("DOM stabilization timed out.")

NOISY_URLS = re.compile(r"(google-analytics|gtm|segment|mixpanel|amplitude|hotjar|sentry|datadog|clarity)", re.I)

async def wait_for_network_quiet(
    page,
    *,
    max_inflight: int = 1,
    quiet_ms: int = 500,
    timeout_ms: int = 15000,
    ignore_patterns: Iterable[str] | None = None,
):
    """
    Wait until network is 'quiet':
      - non-noise requests stay <= max_inflight for quiet_ms continuously.
    Logs the elapsed time at INFO and raises TimeoutError on failure.
    """
    ignore_patterns = tuple((p.lower() for p in (ignore_patterns or ())))

    inflight = 0
    quiet_event = asyncio.Event()
    quiet_timer = None

    def is_noise(url: str) -> bool:
        u = (url or "").lower()
        if NOISY_URLS.search(u):
            return True
        # Ignore long-lived connections explicitly
        # (resource_type is available on the request object, but we only have URL here;
        # websockets/SSE often include these substrings—extend as needed)
        if any(k in u for k in ("/ws", "eventsource", "sse")):
            return True
        if any(p in u for p in ignore_patterns):
            return True
        return False

    loop = asyncio.get_event_loop()

    def cancel_timer():
        nonlocal quiet_timer
        if quiet_timer and not quiet_timer.cancelled():
            quiet_timer.cancel()
        quiet_timer = None

    def arm_timer_if_quiet():
        nonlocal quiet_timer
        if inflight <= max_inflight:
            cancel_timer()
            quiet_timer = loop.call_later(quiet_ms / 1000, quiet_event.set)
        else:
            quiet_event.clear()
            cancel_timer()

    def on_request(req):
        nonlocal inflight
        try:
            if not is_noise(req.url):
                inflight += 1
                quiet_event.clear()
                cancel_timer()
        except Exception:
            # Never let handlers crash
            pass

    def on_done(req):
        nonlocal inflight
        try:
            if not is_noise(req.url):
                inflight = max(0, inflight - 1)
                arm_timer_if_quiet()
        except Exception:
            pass

    # Attach listeners
    page.on("request", on_request)
    page.on("requestfinished", on_done)
    page.on("requestfailed", on_done)

    # Helper to detach listeners (supports both modern & old Playwright)
    def _remove_listener(event_name, handler):
        if hasattr(page, "remove_listener"):
            page.remove_listener(event_name, handler)
        elif hasattr(page, "off"):  # unlikely in Python, but safe fallback
            page.off(event_name, handler)
        else:
            # Last-resort: do nothing; GC will clean up when page closes
            pass

    start = time.monotonic()
    try:
        # Kick off the initial quiet timer based on current inflight (0)
        arm_timer_if_quiet()
        await asyncio.wait_for(quiet_event.wait(), timeout=timeout_ms / 1000)
        elapsed_ms = int((time.monotonic() - start) * 1000)
        logger.debug(f"wait_for_network_quiet: network quiet after {elapsed_ms} ms (<= {max_inflight} in-flight for {quiet_ms} ms)")
    except asyncio.TimeoutError as e:
        elapsed_ms = int((time.monotonic() - start) * 1000)
        logger.error(
            f"wait_for_network_quiet: timeout after {elapsed_ms} ms "
            f"(in-flight={inflight}, threshold={max_inflight}, quiet_ms={quiet_ms}, timeout_ms={timeout_ms})"
        )
        raise TimeoutError("network-idle timeout") from e
    finally:
        cancel_timer()
        _remove_listener("request", on_request)
        _remove_listener("requestfinished", on_done)
        _remove_listener("requestfailed", on_done)


async def start_expansion_watch(control_locator):
    """
    Start a minimal DOM/ARIA watch in the locator's frame.
    - Records baseline DOM element count (top document only, minus __surfari_reasoning_box__).
    - Starts a MutationObserver just to flip popup/overlay flags when nodes appear.
    - Finds the nearest ARIA element (self -> closest ancestor -> container query)
      and records its baseline aria-expanded / aria-haspopup values.
    Returns immediately.
    """
    return await control_locator.evaluate("""
    (node) => {
      // Clean any previous run in this frame
      try { window.__surfariWatch?.cleanup?.(); } catch(_) {}

      const out = {
        domCountStart: 0,
        popup: false,
        overlay: false,
        startedAt: Date.now(),
        // ARIA baseline (filled below)
        ariaFound: false,
        ariaStrategy: null,
        ariaTag: null,
        ariaId: null,
        ariaClass: null,
        ariaExpandedBefore: null,
        ariaHaspopupBefore: null
      };

      // EXACT domElementCount the user provided
      const domElementCount = () => {
        try {
          return document.querySelectorAll('*').length
               - document.querySelectorAll('#__surfari_reasoning_box__').length;
        } catch(_) { return 0; }
      };

      // Detect popup/overlay on newly added nodes
      const mark = (n) => {
        if (!n || n.nodeType !== 1) return;
        const role = n.getAttribute?.('role') || '';
        if (role === 'dialog' || role === 'menu' || role === 'listbox') out.popup = true;
        const st = getComputedStyle(n);
        if (st.position === 'fixed' && parseFloat(st.opacity) > 0.01 &&
            (parseInt(st.width)||0) > 200 && (parseInt(st.height)||0) > 100) {
          out.overlay = true;
        }
      };

      // Minimal MutationObserver (top document only)
      let mo = null;
      try {
        mo = new MutationObserver(muts => {
          for (const m of muts) if (m.type === 'childList') {
            for (const n of m.addedNodes || []) mark(n);
          }
        });
        mo.observe(document.documentElement, { subtree: true, childList: true });
      } catch(_) {}

      // Find nearest ARIA element relative to control
      const norm = (v) => (v == null ? null : String(v).toLowerCase());
      const hasAria = (el) => el && (el.hasAttribute('aria-expanded') || el.hasAttribute('aria-haspopup'));

      let ariaTarget = null;
      let strategy = "self";
      if (hasAria(node)) {
        ariaTarget = node;
        strategy = "self";
      } else {
        ariaTarget = node.closest?.('[aria-expanded],[aria-haspopup]') || null;
        if (ariaTarget) {
          strategy = "closest-ancestor";
        } else {
          const container = node.closest?.('[id],[class]') || node.parentElement || document.body;
          ariaTarget = container.querySelector?.('[aria-expanded],[aria-haspopup]') || null;
          if (ariaTarget) strategy = "container-query";
        }
      }

      if (ariaTarget) {
        out.ariaFound = true;
        out.ariaStrategy = strategy;
        out.ariaTag = ariaTarget.tagName || null;
        out.ariaId = ariaTarget.id || null;
        out.ariaClass = ariaTarget.className || null;
        out.ariaExpandedBefore = norm(ariaTarget.getAttribute?.('aria-expanded'));
        out.ariaHaspopupBefore = norm(ariaTarget.getAttribute?.('aria-haspopup'));
      }

      out.domCountStart = domElementCount();

      window.__surfariWatch = {
        out,
        ariaTarget,
        cleanup: () => { try { mo && mo.disconnect(); } catch(_) {} }
      };

      // Return a tiny baseline (optional)
      return { started: true, baseline: out };
    }
    """)

async def finish_expansion_watch(control_locator):
    """
    Stop the watch started by start_expansion_watch(), take final snapshot, compare ARIA,
    and return {"safe": bool, "reason": str, "metrics": {...}}.
    - DOM counts & popup/overlay reflect only the top document (no iframes).
    - ARIA check looks at the same element found at start; we report if aria-expanded flipped
      from "false" to "true".
    """
    obs = await control_locator.evaluate("""
    () => {
      const sess = window.__surfariWatch;
      if (!sess || !sess.out) return { error: "no-session" };
      const out = sess.out;
      const ariaTarget = sess.ariaTarget || null;

      const domElementCount = () => {
        try {
          return document.querySelectorAll('*').length
               - document.querySelectorAll('#__surfari_reasoning_box__').length;
        } catch(_) { return 0; }
      };
      const norm = (v) => (v == null ? null : String(v).toLowerCase());

      const domCountEnd = domElementCount();
      const durationMs = Date.now() - (out.startedAt || Date.now());

      // ARIA after
      const stillThere = !!(ariaTarget && ariaTarget.isConnected);
      const ariaExpandedAfter = norm(stillThere ? ariaTarget.getAttribute('aria-expanded') : null);
      const ariaHaspopupAfter = norm(stillThere ? ariaTarget.getAttribute('aria-haspopup') : null);

      // Done; clean up and clear state
      try { sess.cleanup?.(); } catch(_) {}
      try { delete window.__surfariWatch; } catch(_) {}

      return {
        // DOM metrics
        domCountStart: out.domCountStart,
        domCountEnd,
        netDomDelta: domCountEnd - out.domCountStart,
        popup: !!out.popup,
        overlay: !!out.overlay,
        durationMs,

        // ARIA comparison
        ariaFound: !!out.ariaFound,
        ariaStrategy: out.ariaStrategy,
        ariaDetachedAfter: !!out.ariaFound && !stillThere,
        ariaTag: out.ariaTag, ariaId: out.ariaId, ariaClass: out.ariaClass,
        ariaExpandedBefore: out.ariaExpandedBefore,
        ariaExpandedAfter,
        ariaHaspopupBefore: out.ariaHaspopupBefore,
        ariaHaspopupAfter,

        // Booleans
        ariaChanged: out.ariaExpandedBefore !== ariaExpandedAfter,
        ariaFlippedFalseToTrue: (out.ariaExpandedBefore === "false" && ariaExpandedAfter === "true"),
      };
    }
    """)

    if obs and obs.get("error") == "no-session":
        return {"safe": False, "reason": "observer session missing (navigation or not started)", "metrics": {}}

    # Build metrics
    metrics = {
        # DOM
        "domCountStart":        int(obs.get("domCountStart", 0)),
        "domCountEnd":          int(obs.get("domCountEnd", 0)),
        "netDomDelta":          int(obs.get("netDomDelta", 0)),
        "popup":                bool(obs.get("popup", False)),
        "overlay":              bool(obs.get("overlay", False)),
        "durationMs":           int(obs.get("durationMs", 0)),

        # ARIA
        "ariaFound":            bool(obs.get("ariaFound", False)),
        "ariaStrategy":         obs.get("ariaStrategy"),
        "ariaDetachedAfter":    bool(obs.get("ariaDetachedAfter", False)),
        "ariaTag":              obs.get("ariaTag"),
        "ariaId":               obs.get("ariaId"),
        "ariaClass":            obs.get("ariaClass"),
        "ariaExpandedBefore":   obs.get("ariaExpandedBefore"),
        "ariaExpandedAfter":    obs.get("ariaExpandedAfter"),
        "ariaHaspopupBefore":   obs.get("ariaHaspopupBefore"),
        "ariaHaspopupAfter":    obs.get("ariaHaspopupAfter"),
        "ariaChanged":          bool(obs.get("ariaChanged", False)),
        "ariaFlippedFalseToTrue": bool(obs.get("ariaFlippedFalseToTrue", False)),
    }

    # Minimal evaluation (tune thresholds as desired)
    popup_opened   = metrics["popup"] or metrics["overlay"]
    big_dom_change = abs(metrics["netDomDelta"]) > 40
    aria_opened    = metrics["ariaFlippedFalseToTrue"]

    if popup_opened or big_dom_change or aria_opened:
        reasons = []
        if popup_opened:   reasons.append("popup/overlay added")
        if big_dom_change: reasons.append("large DOM change detected")
        if aria_opened:    reasons.append("aria-expanded changed from false to true")
        return {"safe": False, "reason": " / ".join(reasons), "metrics": metrics}

    return {"safe": True, "reason": "safe", "metrics": metrics}


async def take_actions(page, locator_actions, num_steps=1, reasoning=None) -> list[dict]:
    """
    Execute a series of actions on a page using Playwright.

    Args:
        locator_actions: a JSON string or object containing a list of locator actions to perform.
            Each locator action should be a dictionary with the following keys:
            - "action": The action to perform (e.g., "click", "fill", "select").
            - "locator": The locator string or locator object for the element to interact with.
            - "value": The value to fill or select (optional).

    Returns:
        a list of locator actions with their execution results.
    """    
    logger.debug(f"Number of steps to perform: {num_steps}")
    
    if isinstance(locator_actions, str):
        locator_actions = remove_unescaped_control_characters(locator_actions)
        locator_actions = json.loads(locator_actions)

    # if somehow action is not a list but a dict, convert it to a list
    if not isinstance(locator_actions, list):     
        locator_actions = [locator_actions]
        
    # extract the list of just locators from locator_actions
    locators = [action.get("locator") for action in locator_actions if "locator" in action]
    # try:
    #     await highlight_elements(page=page, elements=locators, color="red", duration=1000)
    # except Exception as e:
    #     logger.error(f"Error highlighting elements: {e}")
    
    skip_subsequent_actions = False
    for i, locator_action in enumerate(locator_actions, start=1):
        action_name = f"locator_action {i}"
        if skip_subsequent_actions:
            logger.debug(f"{action_name}: Skipping subsequent actions due to previous action.")
            locator_action["result"] = "Wait: The last successful action caused the page to show/hide elements. You need to re-evaluate based on the current page content."
            break
        locator_action_copy = dict(locator_action) # shallow copy for logging
        locator_action_copy.pop("value", None)  # Remove value if it exists for logging  
        logger.sensitive(f"Examining and performing {action_name}: {locator_action_copy}")

        is_expandable_element = locator_action.get("is_expandable_element", False)
        if is_expandable_element:
            skip_subsequent_actions = True
        
        action = locator_action.get("action")
        locator = locator_action.get("locator")
        value = locator_action.get("value")
        target = locator_action.get("target")
        if not action:
            logger.warning(f"{action_name}: No action provided. Skipping.")
            locator_action["result"] = "Error: No action provided"
            continue
        
        if not locator:
            logger.warning(f"{action_name}: No locator provided. Skipping.")
            locator_action["result"] = "Error: No locator provided"
            continue

        if action in ("fill", "select") and not value:
            logger.warning(f"{action_name}: No value provided for {action}.")
            locator_action["result"] = "Error: No value provided"
            continue

        try:
            if isinstance(locator, str):
                element = eval(locator)
            elif isinstance(locator, Locator):
                element = locator
            else:
                logger.warning(f"{action_name}: Invalid locator type. Skipping.")
                locator_action["result"] = "Error: Invalid locator type"
                continue
            
            element_count = await element.count()            
            if element_count == 0:
                logger.warning(f"{action_name}: Element not found. Skipping.")
                locator_action["result"] = "Error: Element not found"
                continue

            # If multiple elements found, look for the first visible one
            if element_count > 1:
                visible_element = None
                for i in range(element_count):
                    current_element = element.nth(i)
                    if await current_element.is_visible():
                        visible_element = current_element
                        break
                
                if visible_element:
                    element = visible_element
                else:
                    logger.info(f"{action_name}: Multiple elements found but none visible, using first element")
                    element = element.first()

        except Exception as e:
            logger.error(f"{action_name}: Error eval-ing locator: {e}")
            locator_action["result"] = f"Error: Invalid locator: {e}"
            continue

        element_disabled = await element.is_disabled()
        if element_disabled:
            logger.warning(f"{action_name}: Element is disabled. Skipping.")
            locator_action["result"] = "Error: Element is currently disabled. You should try something else"
            continue
        
        show_reasoning_box_duration = config.CONFIG["app"].get("show_reasoning_box_duration", 2000)    
        try:
            try:
                logger.debug(f"{action_name}: Attempting to scroll element into view and move mouse")
                await element.scroll_into_view_if_needed(timeout=2000)
                await element.wait_for(timeout=2000, state="visible")
                await move_mouse_to(element)
                if config.CONFIG["app"].get("show_reasoning_box", True) and reasoning:
                    await show_reasoning_box(page, element, reasoning, show_reasoning_box_duration)
                logger.debug("Successfully scrolled element into view, element is visible, mouse moved to element, and reasoning box shown")
            except Exception as e:
                logger.error(f"{action_name}: Will force after encountering error preparing for action: {e}")
                
            if action == "click":
                if True: # for now, always do this
                    try:
                        await element.click(timeout=2000, force=True)
                        logger.debug(f"{action_name}: Clicked element using Playwright click")
                    except Exception as e:
                        logger.error(f"{action_name}: Retry with direct JS evaluate after error force clicking element: {e}")
                        y_pos = await element.evaluate("el => el.getBoundingClientRect().top + window.scrollY", timeout=2000)
                        await page.evaluate(f"() => window.scrollTo(0, {int(y_pos) - 100})")
                        await element.evaluate("""
                        el => {
                            const event = new MouseEvent('click', {
                                bubbles: true,
                                cancelable: true,
                                view: window
                            });
                            el.dispatchEvent(event);
                        }
                        """, timeout=2000)
                        logger.debug(f"{action_name}: Clicked element using JS evaluate")                        

            elif action == "fill":
                await element.click(timeout=2000, force=True)
                tag_name = await element.evaluate("el => el.tagName", timeout=2000)
                if tag_name and tag_name.lower() == "td":
                    # customization for handsontable                    
                    await element.dblclick(timeout=2000, force=True)
                    logger.debug(f"{action_name}: Double clicked td element to edit")
                    input_locator = page.locator('textarea.handsontableInput[data-hot-input]')
                    await input_locator.fill(value, timeout=2000, force=True)
                else:
                    await start_expansion_watch(element)
                    type = await element.evaluate("el => el.type", timeout=2000)    
                    if type and type.lower() == "number":
                        # If the input is a number, fill it with the value
                        await element.fill(value, timeout=2000, force=True)
                    else:
                        await element.clear(timeout=2000, force=True) 
                        await page.wait_for_timeout(300)  # Wait for .3 second before typing           
                        # For other types, use press_sequentially
                        # await element.fill(value, timeout=2000, force=True)
                        await element.press_sequentially(value, delay=50)
                    res = await finish_expansion_watch(element)
                    if not res["safe"]:
                        # Log or handle warning
                        logger.info(f"{action_name}: Filling {target} might have caused page to change: {res['reason']}. Metrics: {json.dumps(res['metrics'])}")
                        locator_action["result"] = f"Success with note: filling {target} caused the page layout to change, potentially to show matches or suggestions. If they appear, click to select the match."
                        break
                    else:
                        logger.debug(f"{action_name}: after filling in data: {json.dumps(res['metrics'])}")

            elif action == "select":
                await element.select_option(value, timeout=10000, force=True)
            elif action == "check":
                try:
                    await element.check(timeout=1000, force=True)
                except Exception as e:
                    await element.evaluate("""
                    el => {
                        let match = el.closest('mat-checkbox, [role="checkbox"], label, [role="radio"], input[type="checkbox"], input[type="radio"]');
                        // console.log('[closest check]', match);
                        if (match) match.click();
                    }
                    """)

            elif action == "uncheck":
                try:
                    await element.uncheck(timeout=1000, force=True)
                except Exception as e:
                    await element.evaluate("""
                    el => {
                        let match = el.closest('mat-checkbox, [role="checkbox"], label, [role="radio"], input[type="checkbox"], input[type="radio"]');
                        // console.log('[closest check]', match);
                        if (match) match.click();
                    }
                    """)                                        
            elif action == "dbclick":
                await element.dblclick(timeout=1000, force=True)
            else:
                logger.warning(f"{action_name}: Unsupported action: {action}. Skipping.")
                locator_action["result"] = f"Error: Unsupported action: {action}"
                continue

            locator_action["result"] = "success"
            await page.wait_for_timeout(show_reasoning_box_duration + 100)  # Wait for show_reasoning_box_duration after each action
        except Exception as e:
            logger.error(f"{action_name}: Error performing action: {e}")
            locator_action["result"] = f"Error: failed to perform action: {e}"
        
        if i == num_steps:
            break
      
    return locator_actions

async def hideOrShowWindow(context=None, page=None):
    run_in_background = (
        config.CONFIG["app"]["run_in_background"]
        or os.getenv("SURFARI_IN_BACKGROUND", "False").lower() == "true"
    )
    logger.debug(f"run_in_background: {run_in_background}")
    if not run_in_background:
        return    
     
    if context and page:
        logger.info("Minimizing window after launch with CDP")
        client = await context.new_cdp_session(page)
        window_info = await client.send("Browser.getWindowForTarget")
        window_id = window_info["windowId"]
        await client.send(
            "Browser.setWindowBounds",
            {
                "windowId": window_id,
                "bounds": {
                    "windowState": "minimized",
                },
            },
        )

async def move_mouse_to(locator):
    if not await locator.is_visible():
        logger.error(f"Can't move mouse to locator as it is not visible: {locator}")
        return
    
    box = await locator.bounding_box(timeout=2000)
    if box:
        x = box["x"] + box["width"] * random.uniform(0, 2) / 2
        y = box["y"] + box["height"] * random.uniform(0, 2) / 2
        await locator.page.mouse.move(x, y, steps=int(random.uniform(0, 5)) + 1)


async def scroll_main_scrollable_down_and_up(page, no_of_scrolls=10) -> bool:
    count = 0
    scrolled = scrolled_started = await scroll_main_scrollable(page) 
    while scrolled and count < no_of_scrolls:
        scrolled = await scroll_main_scrollable(page) 
        count += 1
    if scrolled_started:
        await page.wait_for_timeout(1000)  # Wait for 1 second after scrolling down
        await scroll_main_scrollable(page, to_top=True)
                    
async def scroll_main_scrollable(page, to_top: bool = False) -> bool:
    locator, scrollable_element = await get_main_scrollable_locator(page)
    if not locator:
        logger.warning("No scrollable element found.")
        return False

    logger.info(f"Scrolling {'to top' if to_top else 'to bottom'} of: {scrollable_element}")

    before = await locator.evaluate("el => el.scrollTop")

    scroll_script = """
        (el, toTop) => {
            const target = toTop ? 0 : el.scrollHeight - el.clientHeight;
            el.scrollTo({ top: target, behavior: 'smooth' });
            console.log('[scroll]', `Smooth scroll to: ${target}`);
        }
    """
    await locator.evaluate(scroll_script, to_top)

    await page.wait_for_timeout(50)

    after = await locator.evaluate("el => el.scrollTop")
    logger.info(f"[scroll] Scrolled from {before} to {after}")

    return after != before if to_top else after > before
    
    
async def list_scrollable_elements(page):
    return await page.evaluate("""
    () => {
        function isScrollable(el) {
            const style = getComputedStyle(el);
            return (style.overflowY === 'auto' || style.overflowY === 'scroll') &&
                   el.scrollHeight > el.clientHeight;
        }

        const scrollables = [];
        const walker = document.createTreeWalker(document.body, NodeFilter.SHOW_ELEMENT);
        let node = walker.nextNode();
        while (node) {
            if (isScrollable(node)) {
                scrollables.push({
                    tag: node.tagName,
                    id: node.id || null,
                    class: Array.from(node.classList),
                    scrollHeight: node.scrollHeight,
                    clientHeight: node.clientHeight
                });
            }
            node = walker.nextNode();
        }

        // Also consider document-level scrolling
        const doc = document.scrollingElement || document.body;
        if (doc.scrollHeight > doc.clientHeight) {
            scrollables.push({
                tag: doc.tagName,
                id: doc.id || null,
                class: Array.from(doc.classList),
                scrollHeight: doc.scrollHeight,
                clientHeight: doc.clientHeight
            });
        }
        // Sort by scrollHeight descending
        scrollables.sort((a, b) => b.scrollHeight - a.scrollHeight);
        return scrollables;
    }
    """)
    
def css_escape(s: str) -> str:
    # Escape special characters in CSS class names
    return re.sub(r'([^\w-])', lambda m: "\\" + m.group(1), s)

async def get_main_scrollable_locator(page) -> Tuple[Locator, Dict]:
    scrollables = await list_scrollable_elements(page)
    locator = None
    for i, scrollable_element in enumerate(scrollables):
        tag = scrollable_element["tag"]
        id_ = scrollable_element["id"]
        classes = scrollable_element["class"]

        logger.info(f"Scrollable element {i}: {scrollable_element}")

        # Try to create locator
        if id_:
            locator = page.locator(f"#{id_}")
        elif classes:
            escaped_classes = [css_escape(cls) for cls in classes]
            class_selector = "." + ".".join(escaped_classes)
            locator = page.locator(f"{tag}{class_selector}")
        else:
            locator = page.locator(tag)
        # await highlight_elements(page, [locator])
        return locator, scrollable_element # Only get the main (largest) one
    return None, {}

async def highlight_elements(page, elements, color="red", duration=500):
    """Highlight elements with a colored outline"""
    for element in elements:
        await element.evaluate(f"el => el.style.outline = '3px solid {color}'")
    await page.wait_for_timeout(duration)
    for element in elements:
        await element.evaluate("el => el.style.outline = ''")

async def show_reasoning_box(page, locator_or_box = None, reasoning: str = "", show_reasoning_box_duration: int = 2000):
    """
    Show a floating reasoning box next to an element (locator), 
    at a precomputed bounding box, or centered if None.

    Args:
        page: Playwright page object
        locator_or_box: Playwright Locator, dict with {x,y,width,height}, or None
        reasoning: Text to display
        show_reasoning_box_duration: Duration in ms before auto-remove
    """
    # Determine box
    if locator_or_box is None:
        # No locator — box will be handled inside page.evaluate (center placement)
        box = None
    elif hasattr(locator_or_box, "bounding_box"):
        box = await locator_or_box.bounding_box(timeout=2000)
    else:
        box = locator_or_box
    try:
        await page.evaluate(
            """({ box, reasoning, timeoutMs }) => {
                const MARGIN = 16;
                const PADDING = 8;
                const vw = window.innerWidth;
                const vh = window.innerHeight;

                // Remove any existing box
                document.getElementById("__surfari_reasoning_box__")?.remove();

                const div = document.createElement("div");
                div.id = "__surfari_reasoning_box__";
                div.textContent = reasoning;

                Object.assign(div.style, {
                    position: "fixed",
                    background: "#f5f5f5",
                    color: "black",
                    fontSize: "16px",
                    fontWeight: "bold",
                    fontFamily: "Arial, Helvetica, sans-serif",
                    lineHeight: "1.4",
                    WebkitFontSmoothing: "antialiased",
                    padding: "6px 8px",
                    border: "1px solid black",
                    borderRadius: "4px",
                    whiteSpace: "pre-wrap",
                    wordBreak: "break-word",
                    maxWidth: "300px",
                    zIndex: 999999,
                    boxShadow: "0px 0px 5px rgba(0,0,0,0.3)",
                    pointerEvents: "none", // don’t block clicks
                    visibility: "hidden",  // measure first
                    left: "0px",
                    top: "0px",
                });

                document.body.appendChild(div);

                const bw = div.offsetWidth;
                const bh = div.offsetHeight;

                let left, top;

                if (!box) {
                    // Center in viewport
                    left = (vw - bw) / 2;
                    top = (vh - bh) / 2;
                } else {
                    // Prefer right of element
                    left = box.x + box.width + MARGIN;
                    if (left + bw > vw - PADDING) {
                        left = box.x - MARGIN - bw;
                    }
                    left = Math.max(PADDING, Math.min(left, vw - PADDING - bw));

                    // Align vertically with element
                    top = box.y;
                    top = Math.max(PADDING, Math.min(top, vh - PADDING - bh));
                }

                div.style.left = `${left}px`;
                div.style.top  = `${top}px`;
                div.style.visibility = "visible";

                setTimeout(() => div.remove(), timeoutMs);
            }""",
            {"box": box, "reasoning": reasoning, "timeoutMs": show_reasoning_box_duration}
        )
    except Exception as e:
        logger.warning(f"Error showing reasoning box: {e}")

async def inject_control_bar(page, message: str = ""):
    js_code = f"""
    (() => {{
        const controlBar = document.createElement('div');
        controlBar.id = "__surfari_control_bar__";
        controlBar.style.position = 'fixed';
        controlBar.style.bottom = '0';
        controlBar.style.left = '0';
        controlBar.style.right = '0';
        controlBar.style.zIndex = '9999';
        controlBar.style.color = 'black';
        controlBar.style.padding = '10px';
        controlBar.style.fontSize = '14px';
        controlBar.style.display = 'flex';
        controlBar.style.alignItems = 'center';
        controlBar.style.backgroundColor = 'lightgray';
        controlBar.style.fontFamily = 'Arial, sans-serif';
        controlBar.style.boxShadow = '0px -2px 5px rgba(0,0,0,0.2)';
        
        const statusContainer = document.createElement('div');
        statusContainer.style.display = 'flex';
        statusContainer.style.alignItems = 'center';

        const messageSpan = document.createElement('span');
        messageSpan.textContent = {message!r};
        messageSpan.style.fontSize = '16px';
        messageSpan.style.fontWeight = 'bold';
        messageSpan.style.color = '#333';
        messageSpan.style.marginRight = '24px';
        statusContainer.appendChild(messageSpan);

        const toggleButton = document.createElement('button');
        toggleButton.textContent = 'Toggle Mode';
        toggleButton.style.marginLeft = 'auto';
        toggleButton.style.padding = '5px 12px';
        toggleButton.style.border = 'none';
        toggleButton.style.borderRadius = '4px';
        toggleButton.style.backgroundColor = '#555';
        toggleButton.style.color = 'white';
        toggleButton.style.cursor = 'pointer';
        
        controlBar.appendChild(statusContainer);
        controlBar.appendChild(toggleButton);
        document.body.appendChild(controlBar);

        window.surfariMode = false;

        const updateUI = (enabled) => {{
            toggleButton.textContent = enabled ? 'Switch to Manual' : 'Continue to Automation';
            controlBar.style.backgroundColor = enabled ? 'lightgreen' : 'gold';
        }};

        toggleButton.onclick = () => {{
            window.surfariMode = !window.surfariMode;
            updateUI(window.surfariMode);
        }};

        document.addEventListener('submit', (e) => {{
            if (!window.surfariMode) {{
                window.surfariMode = true;
                updateUI(true);
            }}
        }}, true);

        updateUI(window.surfariMode);
    }})();
    """
    await page.evaluate(js_code)

async def remove_control_bar(page):
    await page.evaluate("""
        const bar = document.getElementById("__surfari_control_bar__");
        if (bar) bar.remove();
    """)





"""
Minimal sample tools for Surfari's tool-call framework.

- Pass the callables directly to your normalizers:
    _normalize_tools_for_openai([report_account_details, report_investment_positions])
    _normalize_tools_for_gemini([report_account_details, report_investment_positions])

- Each tool:
  * Validates inputs via Pydantic models
  * Returns ONLY: {"ok": bool, "summary": str}
"""

from typing import Any, List, Optional, Dict, Union
from pydantic import BaseModel, Field
from decimal import Decimal
from surfari.model.tool_helper import _ensure_list_of_models

# -------- Models --------
class Account(BaseModel):
    account_name: str = Field(..., description="The name of the account")
    account_num: Optional[str] = Field(None, description="The number of the account")
    account_value: str = Field(..., description="The value/balance of the account")
    account_type: Optional[str] = Field(None, description="The type/category of the account")

class InvestmentPosition(BaseModel):
    symbol: str = Field(..., description="Ticker symbol, use CASH for cash positions")
    name: Optional[str] = Field(None, description="Instrument name")
    quantity: int = Field(..., description="Quantity/Shares")
    price: float = Field(..., description="Unit price (use 1 for CASH)")
    cost_basis: Optional[str] = Field(None, description="Cost basis as text")
    market_value: Optional[str] = Field(None, description="Market value as text")
    day_change_amount: Optional[str] = Field(None, description="Daily change as text")
    day_change_percentage: Optional[str] = Field(None, description="Daily change percentage as text")
    position_type: Optional[str] = Field(None, description="Type/category of position")
    percent_of_holdings: Optional[str] = Field(None, description="Percent of total holdings as text")

# -------- Tools (minimal returns) --------
def report_account_details(accounts: List[Account]) -> Dict[str, Any]:
    """
    Extract account details from scraped text.
    Returns only: {"ok": bool, "summary": str}
    """
    print("Calling function: report_account_details")
    print("Accounts:", accounts)
    valid, invalid = _ensure_list_of_models(accounts, Account)
    summary = f"accounts={len(valid)}; invalid={invalid}"
    return {"ok": True, "summary": summary}

def report_investment_positions(holdings: List[InvestmentPosition]) -> Dict[str, Any]:
    """
    Extract investment holding details from scraped text.
    Returns only: {"ok": bool, "summary": str}
    """
    print("Calling function: report_investment_positions")
    print("Holdings:", holdings)
    valid, invalid = _ensure_list_of_models(holdings, InvestmentPosition)
    summary = f"positions={len(valid)}; invalid={invalid}"
    return {"ok": True, "summary": summary}

tools = [report_account_details, report_investment_positions]



import argparse
import asyncio
import os
import csv
import sys  # Added for system exit

import surfari.util.config as config
from surfari.security.site_credential_manager import SiteCredentialManager
from surfari.util.cdp_browser import ChromiumManager
from surfari.agents.navigation_agent import NavigationAgent

import surfari.util.surfari_logger as surfari_logger
logger = surfari_logger.getLogger(__name__)


def parse_args():
    """Parses command line arguments for the runner."""
    parser = argparse.ArgumentParser(description="Run Surfari navigation task")
    parser.add_argument("-t", "--task_goal", help="The user task to accomplish (ignored if --batch_file is used)")
    parser.add_argument("-u", "--url", help="The URL to navigate to")
    parser.add_argument("-n", "--site_name", help="Optional site name")
    parser.add_argument("-b", "--use_system_chrome", action="store_true", help="Use system-installed Chrome")
    parser.add_argument("-l", "--llm_model", help="Override default LLM model (e.g. llama3:8b, gpt-4)")
    parser.add_argument("-s", "--enable_data_masking", action="store_true", help="Mask sensitive info")
    parser.add_argument("-m", "--multi_action_per_turn", action="store_true", help="Allow multiple actions per turn")
    parser.add_argument("-U", "--username", help="Username to save for the site (used with --password)")
    parser.add_argument("-P", "--password", help="Password to save for the site (used with --username)")
    parser.add_argument(
        "-f", "--batch_file",
        help="Path to CSV batch file with columns: "
             "task_goal,site_name,url,username,password,enable_data_masking,"
             "multi_action_per_turn,record_and_replay,rr_use_parameterization,use_screenshot,save_screenshot"
    )
    parser.add_argument("-R", "--record_and_replay", action="store_true", help="Enable record-and-replay mode")
    parser.add_argument("-p", "--rr_use_parameterization", action="store_true", help="Enable record-and-replay parameterization")
    parser.add_argument("-S", "--use_screenshot", action="store_true", help="Send screenshot to LLM too")
    parser.add_argument("-w", "--save_screenshot", action="store_true", help="Save screenshots to disk")
    parser.add_argument("-c", "--num_of_tabs", type=int, default=10, help="Number of concurrent tabs to open for batch tasks")
    return parser.parse_args()


async def run_single_task(
    task_goal,
    site_name=None,
    url=None,
    model=None,
    enable_data_masking=False,
    multi_action_per_turn=False,
    username=None,
    password=None,
    use_system_chrome=False,
    record_and_replay=False,
    rr_use_parameterization=False,
    use_screenshot=False,
    save_screenshot=False,
):
    """Executes a single navigation task."""
    page = None
    if username and password and site_name and url:
        cred_manager = SiteCredentialManager()
        cred_manager.save_credentials(site_name=site_name, url=url, username=username, password=password)
        logger.info(f"[{site_name}] Credentials saved")

    manager = await ChromiumManager.get_instance(use_system_chrome=use_system_chrome)
    await asyncio.sleep(1)
    page = await manager.get_new_page()
    logger.info("Successfully got a new page")

    nav_agent = NavigationAgent(
        model=model,
        site_name=site_name,
        url=url,
        enable_data_masking=enable_data_masking,
        multi_action_per_turn=multi_action_per_turn,
        record_and_replay=record_and_replay,
        rr_use_parameterization=rr_use_parameterization,
        use_screenshot=use_screenshot,
        save_screenshot=save_screenshot,
    )
    result = await nav_agent.run(page, task_goal=task_goal)
    logger.info(f"[{site_name or url}] Final answer: {result}")

    try:
        if page and not page.is_closed():
            await page.close()
    except Exception:
        pass


async def _worker(semaphore, kwargs):
    async with semaphore:
        await run_single_task(**kwargs)


async def run_batch_csv(csv_path, model, use_system_chrome, num_of_tabs):
    """Runs tasks from a CSV batch file with limited concurrency."""
    tasks = []
    semaphore = asyncio.Semaphore(num_of_tabs)

    with open(csv_path, newline='') as f:
        reader = csv.DictReader(f)
        for line_num, row in enumerate(reader, 1):
            run_flag = row.get("run", "").strip().lower()
            if run_flag not in ("1", "true", "yes"):
                continue

            task_goal = row.get("task_goal", "").strip() or None
            if not task_goal:
                logger.warning(f"[Line {line_num}] Skipped: task_goal is required")
                continue

            site_name = row.get("site_name", "").strip() or "Unknown Site"
            url = row.get("url", "").strip() or None
            username = row.get("username", "").strip() or None
            password = row.get("password", "").strip() or None

            enable_data_masking = row.get("enable_data_masking", "").strip().lower() in ("1", "true", "yes")
            multi_action_per_turn = row.get("multi_action_per_turn", "").strip().lower() in ("1", "true", "yes")
            record_and_replay = row.get("record_and_replay", "").strip().lower() in ("1", "true", "yes")
            rr_use_parameterization = row.get("rr_use_parameterization", "").strip().lower() in ("1", "true", "yes")
            use_screenshot = row.get("use_screenshot", "").strip().lower() in ("1", "true", "yes")
            save_screenshot = row.get("save_screenshot", "").strip().lower() in ("1", "true", "yes")

            kwargs = {
                "task_goal": task_goal,
                "site_name": site_name,
                "url": url,
                "model": model,
                "enable_data_masking": enable_data_masking,
                "multi_action_per_turn": multi_action_per_turn,
                "use_system_chrome": use_system_chrome,
                "record_and_replay": record_and_replay,
                "rr_use_parameterization": rr_use_parameterization,
                "use_screenshot": use_screenshot,
                "save_screenshot": save_screenshot,
            }
            if username:
                kwargs["username"] = username
            if password:
                kwargs["password"] = password

            logger.info(f"[Line {line_num}] Task: {task_goal} | Site: {site_name} | URL: {url}")
            tasks.append(_worker(semaphore, kwargs))

    await asyncio.gather(*tasks)


async def main():
    args = parse_args()
    if args.llm_model:
        logger.info(f"Using custom LLM model: {args.llm_model}")

    try:
        if args.batch_file:
            if not os.path.isfile(args.batch_file):
                raise FileNotFoundError(f"Batch file not found: {args.batch_file}")
            await run_batch_csv(
                csv_path=args.batch_file,
                model=args.llm_model,
                use_system_chrome=args.use_system_chrome,
                num_of_tabs=args.num_of_tabs,
            )
        else:
            await run_single_task(
                task_goal=args.task_goal,
                site_name=args.site_name,
                url=args.url,
                model=args.llm_model,
                enable_data_masking=args.enable_data_masking,
                multi_action_per_turn=args.multi_action_per_turn,
                username=args.username,
                password=args.password,
                use_system_chrome=args.use_system_chrome,
                record_and_replay=args.record_and_replay,
                rr_use_parameterization=args.rr_use_parameterization,
                use_screenshot=args.use_screenshot,
                save_screenshot=args.save_screenshot,
            )
    except Exception:
        logger.critical("Browser was forcefully closed. Stopping all processes.", exc_info=True)
        sys.exit(1)
    finally:
        await ChromiumManager.stop_instance()


if __name__ == "__main__":
    asyncio.run(main())



import os
import sqlite3
from contextlib import asynccontextmanager, contextmanager

import surfari.util.config as config
import surfari.util.surfari_logger as surfari_logger

logger = surfari_logger.getLogger(__name__)

@contextmanager
def get_db_connection_sync():
    path = os.path.join(config.PROJECT_ROOT, "security", "credentials_dev.db")
    if not os.path.exists(path):
        path = os.path.join(config.PROJECT_ROOT, "security", "credentials.db")
    logger.debug(f"Opening DB at {path} (sync)")
    conn = sqlite3.connect(path)
    conn.row_factory = sqlite3.Row  # Enables dict(row)
    
    try:
        yield conn
    finally:
        conn.close()

@asynccontextmanager
async def get_db_connection():
    path = os.path.join(config.PROJECT_ROOT, "security", "credentials_dev.db")
    if not os.path.exists(path):
        path = os.path.join(config.PROJECT_ROOT, "security", "credentials.db")
    logger.debug(f"Opening DB at {path} (async)")
    conn = sqlite3.connect(path)
    conn.row_factory = sqlite3.Row  # Enables dict(row)
    try:
        yield conn
    finally:
        conn.close()



import time
import json
import os
from dotenv import load_dotenv
from typing import Dict, Union, List, Optional, Any, Callable, Mapping
from threading import Lock
from jsonfinder import jsonfinder
import ollama
from openai import OpenAI
from google import genai
from google.genai import types
from anthropic import Anthropic
import base64
import re

from surfari.model.tool_helper import (
    _normalize_tools_for_openai,
    _normalize_tools_for_gemini,
    _extract_openai_tool_calls,
    _extract_gemini_tool_calls,
)
import surfari.util.config as config
import surfari.util.surfari_logger as surfari_logger
logger = surfari_logger.getLogger(__name__)

# Load environment variables once at the module level.
# This ensures that all instances of LLMClient can access them.
env_path = os.path.join(config.PROJECT_ROOT, "security", ".env_dev")
if not os.path.exists(env_path):
    env_path = os.path.join(config.PROJECT_ROOT, "security", ".env")

logger.debug(f"Loading environment variables from {env_path}")
load_dotenv(dotenv_path=env_path)


class TokenStats:
    """
    A class to manage token statistics, ensuring thread-safe updates.
    Each LLMClient instance will have its own TokenStats object.
    """
    def __init__(self):
        self.token_stats = {}
        self.lock = Lock()

    def update_token_stats(self, agent_name, prompt_token_count, candidates_token_count, prompt_token_cached=0):
        """Adds to the token counts for a given agent name."""
        with self.lock:
            if agent_name not in self.token_stats:
                self.token_stats[agent_name] = {
                    "prompt_token_count": 0,
                    "prompt_token_cached": 0,
                    "candidates_token_count": 0,
                    "total_token_count": 0,
                }
            self.token_stats[agent_name]["prompt_token_count"] += prompt_token_count
            self.token_stats[agent_name]["prompt_token_cached"] += prompt_token_cached
            self.token_stats[agent_name]["candidates_token_count"] += candidates_token_count
            self.token_stats[agent_name]["total_token_count"] += prompt_token_count + candidates_token_count

    def get_token_stats(self) -> Dict[str, Dict[str, int]]:
        """Returns a copy of the current token statistics."""
        with self.lock:
            return self.token_stats.copy()

class LLMClient:
    """
    A client class to handle all LLM interactions, including token tracking.
    Each instance maintains its own TokenStats object.
    """
    def __init__(self):
        """
        Initializes the LLMClient with API keys and its own TokenStats instance.
        """
        self.openai_api_key = os.getenv("OPENAI_API_KEY")
        self.gemini_api_key = os.getenv("GEMINI_API_KEY")
        self.anthropic_api_key = os.getenv("ANTHROPIC_API_KEY")
        self.token_stats = TokenStats()

    def _parse_llm_response_to_json(self, response_text: str) -> Union[Dict, List, None]:
        """
        Parses a response string to a JSON object, handling potential formatting issues.
        """
        json_obj = None
        try:
            json_obj = json.loads(response_text)
        except json.JSONDecodeError:
            try:
                results = jsonfinder(response_text)
                for _, _, found_json_obj in results:
                    if found_json_obj is not None:
                        json_obj = found_json_obj
                        break
            except Exception:
                pass
        if not json_obj:
            logger.error(f"Failed to parse JSON from response: {response_text}")
        return json_obj

    def _loads_map(self, s: Any) -> Optional[Mapping[str, Any]]:
        if isinstance(s, str):
            try:
                v = json.loads(s)
                return v if isinstance(v, Mapping) else None
            except Exception:
                return None
        return s if isinstance(s, Mapping) else None

    def _get_history_content_for_gemini(
        self,
        chat_history: List[Dict[str, Any]]
    ) -> List[Union[types.UserContent, types.ModelContent]]:
        contents: List[Union[types.UserContent, types.ModelContent]] = []
        for m in chat_history:
            role = m.get("role")
            content = m.get("content")

            # Assistant: check if content JSON embeds tool_calls
            if role == "assistant":
                parts: List[types.Part] = []
                j = self._loads_map(content)

                # If content has tool_calls, emit functionCall parts
                if isinstance(j, Mapping) and "tool_calls" in j and isinstance(j["tool_calls"], list):
                    for c in j["tool_calls"]:
                        args = c.get("arguments", {}) or {}
                        if isinstance(args, str):
                            args = self._loads_map(args) or {"value": args}
                        elif not isinstance(args, Mapping):
                            args = {"value": args}
                        parts.append(types.Part.from_function_call(
                            name=c["name"],
                            args=args
                        ))
                else:
                    # otherwise, treat as normal text if present
                    if content:
                        parts.append(types.Part.from_text(text=str(content)))

                if parts:
                    contents.append(types.ModelContent(parts=parts))
                continue

            # Tool results → functionResponse parts (Gemini matches by order)
            if role == "tool":
                payload = self._loads_map(content)
                if payload is None:
                    payload = {"value": content}
                parts = [types.Part.from_function_response(
                    name=m.get("name", "tool"),
                    response=payload
                )]
                contents.append(types.UserContent(parts=parts))
                continue

            # Regular user text
            if role == "user":
                contents.append(types.UserContent(parts=[
                    types.Part.from_text(text=str(content or ""))
                ]))
                continue

            # Fallback assistant/model text
            if role in ("assistant", "model") and content:
                contents.append(types.ModelContent(parts=[
                    types.Part.from_text(text=str(content))
                ]))
        # logger.sensitive("Rebuilt Gemini contents: %s", contents)
        return contents


    def _get_history_content_for_openai(self, chat_history: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Convert your internal history into an OpenAI Responses API 'input' list.

        Rules:
        - role=='user'  -> {"role":"user","content": "..."}
        - role=='assistant' with content JSON containing {"tool_calls":[...]}
                -> one {"type":"function_call", name, call_id, arguments} per call
            else regular assistant text -> {"role":"assistant","content": "..."}
        - role=='tool'  -> {"type":"function_call_output","call_id": <match>, "output": "<json str>"}
            (matched to the most recent unmatched function_call, by order; fallback by name)

        Returns:
        A list of items ready for OpenAI Responses API.
        """
        inputs: List[Dict[str, Any]] = []

        for m in chat_history:
            role = m.get("role")
            content = m.get("content")

            # --- USER TEXT ---
            if role == "user":
                inputs.append({"role": "user", "content": str(content or "")})
                continue

            # --- ASSISTANT: tool_calls encoded in content JSON ---
            if role == "assistant":
                j = self._loads_map(content)
                if isinstance(j, Mapping) and isinstance(j.get("tool_calls"), list):
                    for c in j["tool_calls"]:
                        name = c.get("name")
                        if not name:
                            # Skip malformed call
                            continue
                        args = c.get("arguments", {}) or {}
                        if isinstance(args, str):
                            args_map = self._loads_map(args)
                            args = args_map if args_map is not None else {"value": args}
                        elif not isinstance(args, Mapping):
                            args = {"value": args}
                        call_id = c.get("call_id") or c.get("id")
                        # Emit a function_call item
                        inputs.append({
                            "type": "function_call",
                            "name": name,
                            "call_id": call_id,
                            "arguments": json.dumps(args),  # must be a JSON string
                        })
                else:
                    # Regular assistant text (if any)
                    if content:
                        inputs.append({"role": "assistant", "content": str(content)})
                continue

            # --- TOOL RESULTS -> function_call_output ---
            if role == "tool":
                # Determine which pending call this output answers
                payload_map = self._loads_map(content)
                payload_str = json.dumps(payload_map if payload_map is not None else {"value": content})

                call_id: Optional[str] = m.get("call_id")
                if call_id:
                    inputs.append({
                        "type": "function_call_output",
                        "call_id": call_id,
                        "output": payload_str,
                    })
                # If no call_id available, we skip; alternatively, raise/log here.
                continue
        logger.sensitive("Rebuilt OpenAI inputs: %s", json.dumps(inputs, indent=2))
        return inputs


    async def process_prompt_return_json(
        self,
        system_prompt: str = "",
        user_prompt: str = "",
        chat_history: List[Dict[str, str]] = [],
        image_data: Optional[bytes] = None,
        image_format: str = "jpeg",
        tools: List[Callable[..., Any]] = [],
        model: str = "gemini-2.0-flash",
        purpose: str = "navigation",
        site_id: int = 0
    ) -> Union[Dict, List, None]:
        """
        Sends a prompt and returns either:
          - Parsed JSON from text (existing behavior), OR
          - {"tool_calls": [{"name":..., "arguments": {...}, "id": ...?}, ...]}
            when the model decides to call tools (OpenAI or Gemini).
        """
     
        prompt_to_log = system_prompt + json.dumps(chat_history, indent=2) + user_prompt
        await logger.log_text_to_file(site_id, prompt_to_log, purpose, "prompt")

        start = time.time()
        is_open_ai = model.startswith("gpt-") or model == "o3-mini"
        is_gemini_ai = model.startswith("gemini-")
        is_anthropic = model.startswith("claude-")
        is_ollama = model.startswith("deepseek") or model.startswith("qwen") or model.startswith("llama") or model.startswith("gemma")

        messages = []
        messages.append({"role": "system", "content": system_prompt})
        if is_ollama:
            messages += chat_history
        elif is_open_ai:
            messages += self._get_history_content_for_openai(chat_history)
            
        if image_data and is_open_ai:
            b64 = base64.b64encode(image_data).decode("ascii")
            data_url = f"data:image/{image_format};base64,{b64}"
            messages.append({
                "role": "user",
                "content": [
                    {"type": "input_text", "text": user_prompt},
                    {"type": "input_image", "image_url": data_url},
                ],
            })
        else:
            messages.append({"role": "user", "content": user_prompt})

        responsetxt = ""
        tool_calls_payload: Optional[Dict[str, Any]] = None

        if is_open_ai:
            client = OpenAI(api_key=self.openai_api_key)
            if model.startswith("gpt-5"):
                kwargs = dict(model=model, input=messages, reasoning={"effort": "minimal"})
            else:
                kwargs = dict(model=model, input=messages)
                
            if tools:
                kwargs["tools"] = _normalize_tools_for_openai(tools)
            response = client.responses.create(**kwargs)
            logger.sensitive("OpenAI response: %s", response)
            # First, try to extract tool calls
            ocalls = _extract_openai_tool_calls(response)
            if ocalls:
                tool_calls_payload = {"tool_calls": ocalls}
            else:
                # Fallback to plain text
                responsetxt = (getattr(response, "output_text", None) or "").strip()

            usage = getattr(response, "usage", None)
            if usage:
                logger.debug(f"OpenAI response usage: {usage}")
                input_tokens = getattr(usage, "input_tokens", 0)
                cached_tokens = 0
                if hasattr(usage, "input_tokens_details") and usage.input_tokens_details:
                    cached_tokens = getattr(usage.input_tokens_details, "cached_tokens", 0)
                output_tokens = getattr(usage, "output_tokens", 0)
                self.token_stats.update_token_stats(purpose, input_tokens, output_tokens, cached_tokens)

        elif is_gemini_ai:
            client = genai.Client(api_key=self.gemini_api_key)
            userContent = types.UserContent(parts=[types.Part.from_text(text=user_prompt)])
            config_args = {
                "automatic_function_calling": types.AutomaticFunctionCallingConfig(disable=True),
                "system_instruction": system_prompt,
            }
            if tools:
                config_args["tools"] = _normalize_tools_for_gemini(tools)
            if model.startswith("gemini-2.5"):
                config_args["thinking_config"] = types.ThinkingConfig(thinking_budget=0)

            contents = self._get_history_content_for_gemini(chat_history) + [userContent]
            if image_data:
                contents = contents + [types.Part.from_bytes(data=image_data, mime_type=f"image/{image_format}")]

            response = client.models.generate_content(
                model=model,
                config=types.GenerateContentConfig(**config_args),
                contents=contents
            )

            logger.sensitive("Gemini response: %s", response)
            # Prefer tool calls if present
            gcalls = _extract_gemini_tool_calls(response)
            if gcalls:
                tool_calls_payload = {"tool_calls": gcalls}
            else:
                responsetxt = (response.text or "").strip()

            usage = response.usage_metadata
            if usage:
                logger.debug(f"Gemini response usage: {usage}")
                
                self.token_stats.update_token_stats(purpose, response.usage_metadata.prompt_token_count, response.usage_metadata.candidates_token_count)

        elif is_anthropic:
            client = Anthropic(api_key=self.anthropic_api_key)
            response = client.messages.create(
                model=model,
                max_tokens=1024,
                temperature=0.7,
                system=system_prompt,
                messages=chat_history + [{"role": "user", "content": user_prompt}],
            )
            logger.sensitive("Anthropic response: %s", response)
            responsetxt = response.content[0].text.strip() if response.content else ""
            usage = getattr(response, "usage", None)
            if usage:
                logger.debug(f"Anthropic response usage: {usage}")
                self.token_stats.update_token_stats(purpose, getattr(usage, "input_tokens", 0), getattr(usage, "output_tokens", 0))

        elif is_ollama:
            client = ollama.Client()
            response = client.chat(model=model, messages=messages)
            logger.sensitive("Ollama response: %s", response)
            responsetxt = response["message"].get("content")

        else:
            raise ValueError("Invalid model")

        end = time.time()
        logger.info(f"Time taken to call LLM : {end - start:.2f}s")

        # If tools were invoked, return them (caller will execute)
        if tool_calls_payload:
            logger.sensitive(f"LLM tool calls: {tool_calls_payload}")
            await logger.log_text_to_file(site_id, json.dumps(tool_calls_payload, indent=2), purpose, "response")
            return tool_calls_payload

        # Else parse JSON from text
        responsejson = self._parse_llm_response_to_json(responsetxt) if responsetxt else None
        await logger.log_text_to_file(site_id, json.dumps(responsejson, indent=2) if responsejson is not None else "null", purpose, "response")
        return responsejson



PARAMETERIZATION_SYSTEM_PROMPT = """
You are an expert in identifying variable components in natural language task descriptions and transforming them into parameterized templates that can be replayed with substituted values.

Your goal:
1. Identify values in the task description that are likely to change between runs.
2. Replace each with a placeholder in the format `:1`, `:2`, `:3`, etc., starting from `:1` in the order they appear in the sentence.
3. Treat as variables:
   - City names, airport names, street names, or other locations.
   - Dates, months, years.
   - Numeric values (quantities, prices, durations, counts).
   - Names of people, companies, organizations.
   - Product names or model identifiers.
4. Keep all other text exactly the same.
5. Do not remove units, descriptors, or context words (e.g., keep "days later", "USD", "kg", "$"). Don't include them in the variable.
6. Maintain the original sentence structure and punctuation.

Return the result **only** in the following JSON format:

{
  "parameterized_task_desc": "<parameterized task>",
  "variables": {
    ":1": "<original value>",
    ":2": "<original value>",
    ...
  }
}

The `parameterized_task_desc` will be stored in the `parameterized_task_desc` column.
The `variables` object will be stored separately for replay logic and will not overwrite the original chat history.

Examples:

Input:
Find tickets from Boston to Seattle, leaving on August 10, 2025, 10 days later return date, direct flight or at most 1 stop, Find lowest price under $500.

Output:
{
  "parameterized_task_desc": "Find tickets from :1 to :2, leaving on :3 :4, :5, :6 days later return date, direct flight or at most :7 stop, Find lowest price under $:8.",
  "variables": {
    ":1": "Boston",
    ":2": "Seattle",
    ":3": "August",
    ":4": "10", 
    ":5": "2025",
    ":6": "10",
    ":7": "1",
    ":8": "500"
  }
}

Input:
Schedule a meeting with Alice Johnson at Google HQ on March 5, 2026, at 2 PM, for the project kickoff

Output:
{
  "parameterized_task_desc": "Schedule a meeting with :1 at :2 on :3 :4, :5 at :6 for :7",
  "variables": {
    ":1": "Alice Johnson",
    ":2": "Google HQ",
    ":3": "March",
    ":4": "5",
    ":5": "2026",
    ":6": "2 PM",
    ":7": "the project kickoff"
  }
}
    ":4": "2 PM",
    ":5": "the project kickoff"
  }
}

Now, process the following task description and return only the JSON result.
"""


import os
import subprocess
import asyncio
import signal
import platform
from typing import Optional, ClassVar
import threading
import pathlib
from playwright.async_api import async_playwright, BrowserContext, Page

import surfari.util.config as config
import surfari.util.surfari_logger as surfari_logger

logger = surfari_logger.getLogger(__name__)

REMOTE_DEBUGGING_PORT = 9222

screen_width = config.CONFIG["app"].get("browser_width", 1712)
screen_height = config.CONFIG["app"].get("browser_height", 1072)

USER_DATA_DIR = os.path.join(config.PROJECT_ROOT, "playwright_chrome_profile")

# Platform-specific Chrome paths
if platform.system() == "Darwin":
    DEFAULT_CHROME_PATH = "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome"

elif platform.system() == "Windows":
    DEFAULT_CHROME_PATH = r"C:\Program Files (x86)\Google\Chrome\Application\chrome.exe"
    if not os.path.isfile(DEFAULT_CHROME_PATH):
        DEFAULT_CHROME_PATH = r"C:\Program Files\Google\Chrome\Application\chrome.exe"

elif platform.system() == "Linux":
    DEFAULT_CHROME_PATH = "/usr/bin/chromium"
    if not os.path.isfile(DEFAULT_CHROME_PATH):
        DEFAULT_CHROME_PATH = "/usr/bin/chromium-browser"

else:
    raise NotImplementedError("Unsupported platform")


init_script_text = """
(() => {
  // === Patch performance.now() to be consistent ===
  const originalNow = performance.now.bind(performance);
  const startOffset = originalNow();
  performance.now = () => originalNow() - startOffset;

  // === Patch console methods to avoid object getter triggers ===
  const safeConsole = ['log', 'debug', 'info', 'warn', 'error', 'dir'];
  for (const method of safeConsole) {
    const original = console[method];
    console[method] = (...args) => {
      const safeArgs = args.map(arg => {
        if (typeof arg === 'object' && arg !== null) {
          try { return JSON.parse(JSON.stringify(arg)); }
          catch (e) { return '[Object]'; }
        }
        return arg;
      });
      return original.apply(console, safeArgs);
    };
  }

  // === Patch debugger timing trap ===
  let lastDebuggerTime = performance.now();
  Object.defineProperty(window, 'debuggerTrap', {
    get() {
      const now = performance.now();
      const delta = now - lastDebuggerTime;
      lastDebuggerTime = now;
      return delta <= 100;
    }
  });

  // === Force all window.open() to open in same tab ===
  /*
  window.open = (url) => {
    if (url) window.location.href = url;
    return null;
  };

  // === Strip target="_blank" from all links ===
  document.addEventListener('DOMContentLoaded', () => {
    document.querySelectorAll('a[target="_blank"]').forEach(a => a.removeAttribute('target'));
  });
  */
})();
"""

class ChromiumManager:
    _instance: ClassVar[Optional["ChromiumManager"]] = None
    _instance_lock: ClassVar[asyncio.Lock] = asyncio.Lock()

    def __init__(
        self,
        use_system_chrome: bool = False,
        remote_debugging_port: int = REMOTE_DEBUGGING_PORT,
        user_data_dir: str = USER_DATA_DIR,
        logger=logger,
        wait_for_browser_start: float = 2.0,
        shutdown_timeout: float = 10.0,
    ):
        self.use_system_chrome = use_system_chrome
        self.remote_debugging_port = remote_debugging_port
        self.user_data_dir = user_data_dir
        self.logger = logger
        self.wait_for_browser_start = wait_for_browser_start
        self.shutdown_timeout = shutdown_timeout

        self.chrome_process: Optional[subprocess.Popen] = None
        self.playwright = None
        self.browser_context: Optional[BrowserContext] = None

        self._loop = asyncio.get_event_loop()
        self._signals_installed = False
        self.stopped = False
        self.logger.info(f"Browser instance initialized with Screen resolution: {screen_width}x{screen_height}")

    @classmethod
    async def get_instance(cls, use_system_chrome=False) -> "ChromiumManager":
        async with cls._instance_lock:
            if cls._instance is None:
                cls._instance = cls(use_system_chrome=use_system_chrome)
                await cls._instance.start()
        return cls._instance

    @classmethod
    async def stop_instance(cls) -> None:
        async with cls._instance_lock:
            if cls._instance:
                await cls._instance.stop()
                cls._instance = None

    async def get_new_page(self) -> Page:
        if not self.browser_context:
            raise RuntimeError("Browser context not yet initialized or closed")
        page = await self.browser_context.new_page()
        await page.add_init_script(init_script_text)
        self.logger.info("New tab created.")
        return page

    async def __aenter__(self):
        raise RuntimeError("Use ChromiumManager.get_instance() instead of context manager.")

    async def __aexit__(self, *args):
        pass

    async def start(self) -> None:
        self.logger.info("Starting ChromiumManager...")
        await self._install_signal_handlers()
        if not self.running_in_container():
            await self._launch_browser()
        await self._connect_over_cdp()

    async def stop(self) -> None:
        if self.stopped:
            self.logger.info("ChromiumManager already stopped.")
            return
        self.logger.info("Stopping ChromiumManager and marking as stopped.")
        self.stopped = True
        await self._close_browser_context()
        await self._shutdown_browser()

    def running_in_container(self):
        try:
            if pathlib.Path("/.dockerenv").exists():
                return True
            with open("/proc/1/cgroup", "r") as f:
                return any(x in f.read() for x in ["docker", "kubepods", "lxc"])
        except Exception:
            return False

    async def _install_signal_handlers(self) -> None:
        if not self._signals_installed and threading.current_thread() is threading.main_thread():
            def handle_signal(sig):
                self.logger.warning(f"Received signal {sig}, shutting down...")
                asyncio.create_task(self.stop())
            if platform.system() in ("Darwin", "Linux"):
                self._loop.add_signal_handler(signal.SIGINT, lambda: handle_signal("SIGINT"))
                self._loop.add_signal_handler(signal.SIGTERM, lambda: handle_signal("SIGTERM"))
            self._signals_installed = True

    def _build_chrome_args(self, executable_path: str) -> list[str]:
        args = [
            executable_path,
            f"--remote-debugging-port={self.remote_debugging_port}",
            "--remote-debugging-address=localhost",
            "--no-first-run",
            "--no-default-browser-check",
            "--disable-webrtc",
            "--disable-background-networking",
            "--disable-features=WebRtcHideLocalIpsWithMdns",
            "--window-position=0,0",
            f"--window-size={screen_width},{screen_height}",
            "--log-level=3",
            f"--user-data-dir={self.user_data_dir}"
        ]
        if platform.system() == "Linux" and not self.use_system_chrome and os.environ.get("WAYLAND_DISPLAY"):
            args += ["--ozone-platform=wayland", "--ozone-platform-hint=auto"]
        return args

    async def _launch_browser(self) -> None:
        if self.use_system_chrome:
            await self._launch_system_chrome()
        else:
            await self._launch_bundled_chromium()
            
        self.logger.info(f"Chrome process started with PID: {self.chrome_process.pid if self.chrome_process else 'N/A'}")
        await asyncio.sleep(self.wait_for_browser_start)

    async def _launch_system_chrome(self) -> None:
        if not os.path.isfile(DEFAULT_CHROME_PATH):
            raise FileNotFoundError(f"System Chrome not found at '{DEFAULT_CHROME_PATH}'")
        chrome_args = self._build_chrome_args(DEFAULT_CHROME_PATH)
        self.logger.info("Launching system Chrome with args: %s", chrome_args)
        self.chrome_process = await asyncio.create_subprocess_exec(*chrome_args)

    async def _launch_bundled_chromium(self) -> None:
        async with async_playwright() as p:
            chromium_path = p.chromium.executable_path
        if not os.path.isfile(chromium_path):
            self.logger.warning(f"Bundled Chromium not found at '{chromium_path}', using system Chrome instead.")
            await self._launch_system_chrome()
            return
        chrome_args = self._build_chrome_args(chromium_path)
        self.logger.info("Launching bundled Chromium with args: %s", chrome_args)
        self.chrome_process = await asyncio.create_subprocess_exec(*chrome_args)

    async def _connect_over_cdp(self) -> None:
        for attempt in range(1, 4):
            try:
                self.logger.info(f"Attempt {attempt}: Connecting over CDP...")
                self.playwright = await async_playwright().start()
                browser = await self.playwright.chromium.connect_over_cdp(
                    f"http://localhost:{self.remote_debugging_port}"
                )

                contexts = browser.contexts
                if contexts:
                    self.browser_context = contexts[0]
                    self.logger.info("Reusing existing browser context.")
                else:
                    self.browser_context = await browser.new_context()
                    self.logger.info("Created a new browser context.")

                self.browser_context.on(
                    "close",
                    lambda: (
                        self.logger.info("BrowserContext closed."),
                        setattr(self, 'browser_context', None)
                    )
                )
                await self.browser_context.add_init_script(init_script_text)
                self.logger.info("Connected over CDP. BrowserContext is ready.")
                return  # success
            except Exception as e:
                self.logger.error(f"Attempt {attempt} failed to connect over CDP: {e}")
                if attempt == 3:
                    raise
                await asyncio.sleep(3)

    async def _close_browser_context(self) -> None:
        self.logger.info("Closing BrowserContext called.")
        if self.browser_context:
            self.logger.info("Closing BrowserContext...")
            try:
                await self.browser_context.close()
            except Exception as e:
                self.logger.error(f"Error closing BrowserContext: {e}")
            self.browser_context = None

        if self.playwright:
            self.logger.info("Stopping Playwright...")
            await self.playwright.stop()
            self.playwright = None
        # await asyncio.sleep(3)

    async def _shutdown_browser(self) -> None:
        self.logger.info("Shutting down browser called.")
        if self.chrome_process:
            self.logger.info(f"Terminating browser process with PID: {self.chrome_process.pid}")
            try:
                if self.chrome_process.returncode is None:
                    self.logger.info("Waiting for browser to terminate gracefully...")
                    self.chrome_process.terminate()
                    # await asyncio.wait_for(self.chrome_process.wait(), timeout=self.shutdown_timeout)
                    self.logger.info("Browser terminated gracefully.")
                else:
                    self.logger.info("Browser already terminated with exit code: %d", self.chrome_process.returncode)
            except asyncio.TimeoutError:
                self.logger.warning("Force killing browser...")
                self.chrome_process.kill()
            except Exception as e:
                self.logger.error(f"Shutdown error: {e}")
            finally:
                self.chrome_process = None




import re
import time
import asyncio
from typing import Optional, List, Dict, Any
import surfari.util.surfari_logger as surfari_logger
from surfari.agents.tools.google_tools import gmail_search_emails

logger = surfari_logger.getLogger(__name__)

class GmailOTPClientAsync:
    """
    OTP helper built on top of google_tools' high-level functions.
    - No direct OAuth/HTTP here; we rely on google_tools to handle scopes, tokens, retries.
    """

    def __init__(self):
        # retain a spot for future config if needed
        pass

    # ---------- Public API ----------
    async def get_otp_code(
        self,
        from_me: bool = True,
        within_seconds: int = 30,
        retry_interval: int = 10,
        max_retries: int = 6,
        max_results: int = 10,
        include_body_lookup: bool = True,
    ) -> Optional[str]:
        """
        Retry loop to fetch the most recent OTP code.
        - from_me: restrict to messages sent by me
        - within_seconds: only search recent messages (Gmail supports 'after:<unix_ts>')
        - retry_interval: seconds to wait between attempts
        - max_retries: number of attempts
        - max_results: how many messages to fetch per query (most recent first)
        - include_body_lookup: if true, fall back to reading message body/snippet if Subject has no code
        """
        for attempt in range(1, max_retries + 1):
            logger.debug(f"[🔄] OTP fetch attempt {attempt}/{max_retries}…")
            code = await self.get_latest_code(
                from_me=from_me,
                within_seconds=within_seconds,
                max_results=max_results,
                include_body_lookup=include_body_lookup,
            )
            if code:
                return code
            if attempt < max_retries:
                logger.debug(f"[⏳] No OTP yet; sleeping {retry_interval}s before next attempt…")
                await asyncio.sleep(retry_interval)
        return None

    async def get_latest_code(
        self,
        from_me: bool = True,
        within_seconds: int = 600,
        max_results: int = 10,
        include_body_lookup: bool = True,
    ) -> Optional[str]:
        """
        Single-shot attempt to fetch the latest OTP code within the time window.
        Returns the first matching code found (most recent first).
        """
        query = self._build_query(from_me=from_me, within_seconds=within_seconds)
        logger.debug(f"[>] Gmail OTP query: {query}")

        # Use the google_tools wrapper (handles auth and errors)
        resp = await gmail_search_emails(query=query, max_results=max_results)
        if not resp.get("ok"):
            logger.warning(f"[!] gmail_search_emails failed: {resp.get('error')}")
            return None

        messages: List[Dict[str, Any]] = resp.get("messages", []) or []
        if not messages:
            return None

        # Examine recent messages in order; prefer Subject-based code
        for m in messages:
            headers = (m.get("headers") or {})
            subject = (headers.get("Subject") or "").strip()
            snippet = (m.get("snippet") or "").strip()

            # 1) Try subject first (most reliable signal)
            code = self._extract_code_from_subject(subject)
            if code:
                logger.debug(f"[✓] OTP from Subject: {subject!r} -> {code}")
                return code

            # 2) Fall back to snippet (optional)
            if include_body_lookup:
                body_code = self._extract_code_from_text(snippet)
                if body_code:
                    logger.debug(f"[✓] OTP from snippet for msg {m.get('id')}: {body_code}")
                    return body_code

        logger.debug("[!] No OTP found in recent messages.")
        return None

    # ---------- Helpers ----------

    def _build_query(self, *, from_me: bool, within_seconds: int) -> str:
        """
        Build a Gmail search query limiting to recent messages,
        optionally those 'from:me', and nudging for OTP-like subjects.
        """
        since_ts = int(time.time()) - max(0, int(within_seconds))
        # Common patterns for OTP/verification emails—help Google rank the right messages first.
        # We keep it broad: subject filters improve relevance but aren't required.
        subject_hint = '(subject:code OR subject:verification OR subject:passcode OR subject:OTP)'
        base = f"after:{since_ts} label:inbox {subject_hint}"
        return (f"from:me {base}" if from_me else base).strip()

    def _extract_code_from_subject(self, subject: str) -> Optional[str]:
        """
        Extract a code when 'code' (or typical OTP words) appears in the subject.
        Prioritize subjects that contain those words; else return None.
        """
        s = subject.lower()
        if not any(w in s for w in ("code", "otp", "passcode", "verification")):
            return None
        return self._extract_code_from_text(subject)

    def _extract_code_from_text(self, text: str) -> Optional[str]:
        """
        Extract a 4–8 digit code from text. Adjust as needed.
        """
        match = re.search(r"\b(\d{4,8})\b", text or "")
        return match.group(1) if match else None


# Example usage (manual test)
async def _example():
    client = GmailOTPClientAsync()
    code = await client.get_otp_code(from_me=True, within_seconds=300, retry_interval=10, max_retries=6)
    logger.info(f"Latest OTP: {code}")

if __name__ == "__main__":  # pragma: no cover
    asyncio.run(_example())



from dataclasses import dataclass
from typing import Any, Optional
import os
from dotenv import load_dotenv
from pinecone import Pinecone
from surfari.agents.navigation_agent._typing import ResolveInput, ResolveOutput
import surfari.util.config as config
import surfari.util.surfari_logger as surfari_logger

logger = surfari_logger.getLogger(__name__)

# Load env
env_path = os.path.join(config.PROJECT_ROOT, "security", ".env_dev")
if not os.path.exists(env_path):
    env_path = os.path.join(config.PROJECT_ROOT, "security", ".env")
    logger.debug(f"PineconeResolver:Loading environment variables from {env_path}")
load_dotenv(dotenv_path=env_path)

@dataclass
class PineconeManagedEmbedResolver:
    """
    Pinecone resolver using Pinecone-managed embeddings.
    On upsert, include metadata["chunk_text"] with your raw text.
    On query, we ask Pinecone to embed the query text server-side, then search.
    """
    index: str
    embed_model: str = "llama-text-embed-v2"  # must match index managed model
    namespace: Optional[str] = None
    score_threshold: Optional[float] = None
    top_k: int = 3

    _pc: Any = None
    _pc_index: Any = None

    def __post_init__(self) -> None:
        api_key = os.getenv("PINECONE_API_KEY")
        if not api_key:
            raise RuntimeError("Missing PINECONE_API_KEY in your environment.")
        self._pc = Pinecone(api_key=api_key)
        self._pc_index = self._pc.Index(self.index)
        logger.info("Initialized Pinecone index=%r (managed model=%r)", self.index, self.embed_model)


    def resolve(self, inp: ResolveInput) -> ResolveOutput:
        """
        Use Pinecone-managed embeddings via Index.search().
        Expects search() result like:
        {"result": {"hits": [{"_id": "...", "_score": 7.8, "fields": {...}}, ...]}, "usage": {...}}
        """
        query_text = (inp.text or "").strip()
        if not query_text:
            return ResolveOutput(value=None)

        ns = self.namespace or ""

        # Ask Pinecone to embed + search server-side
        results = self._pc_index.search(
            namespace=ns,
            query={
                "inputs": {"text": query_text},
                "top_k": self.top_k,
            },
            # request back fields we stored during upsert
            fields=["chunk_text", "value", "label"],
        )
        # logger.debug("Pinecone search results: %r", results)
        # ---- normalize hits from the response ----
        # handle both dict and possible SDK object shapes
        if isinstance(results, dict):
            result_block = results.get("result") or {}
            hits = result_block.get("hits") or []
        else:
            # SDK object fallback
            try:
                result_block = getattr(results, "result", None) or {}
                hits = getattr(result_block, "hits", None) or []
            except Exception:
                hits = []

        if not hits:
            return ResolveOutput(value=None)

        # score helper (dict with "_score")
        def _score(hit: Any) -> float:
            if isinstance(hit, dict):
                return float(hit.get("_score", 0.0))
            # fallback if SDK object
            return float(getattr(hit, "_score", 0.0))

        best = max(hits, key=_score)
        best_score = _score(best)

        if self.score_threshold is not None and best_score < self.score_threshold:
            logger.debug("Best score %.4f below threshold %.4f", best_score, self.score_threshold)
            return ResolveOutput(value=None)

        # fields come back under "fields"
        fields = best.get("fields", {}) if isinstance(best, dict) else getattr(best, "fields", {}) or {}
        value = None
        if isinstance(fields, dict):
            # prefer an explicit value; fall back to other fields you stored
            value = fields.get("value") or fields.get("chunk_text") or fields.get("label")

        # last resort: try metadata (older patterns)
        if value is None:
            md = best.get("metadata", {}) if isinstance(best, dict) else getattr(best, "metadata", {}) or {}
            if isinstance(md, dict):
                value = md.get("value") or md.get("text") or md.get("chunk_text")
                
        logger.debug("Pinecone resolve: query=%r, best_score=%.4f, value=%r", query_text, best_score, value)
        return ResolveOutput(value=str(value) if value is not None else None)


def main() -> None:
    api_key = os.getenv("PINECONE_API_KEY")
    if not api_key:
        raise RuntimeError("Missing PINECONE_API_KEY in your environment.")

    pc = Pinecone(api_key=api_key)

    resolver_cfg = {
        "target": "surfari.agents.navigation_agent._pinecone_resolver:PineconeManagedEmbedResolver",
        "params": {
            "index": "surfari-index2",
            "namespace": "default",                    
            "top_k": 3,
            "score_threshold": 0.5
        }
    }
    if "value_resolver" in config.CONFIG and config.CONFIG["value_resolver"]:
        if config.CONFIG["value_resolver"].get("target", "") == "surfari.agents.navigation_agent._pinecone_resolver:PineconeManagedEmbedResolver":
            resolver_cfg = config.CONFIG["value_resolver"]
    
    params = resolver_cfg.get("params", {})  
    
    index_name = params.get("index", "surfari-index")
    region = params.get("region", "us-east-1")
    cloud = params.get("cloud", "aws")
    embed_model = params.get("embed_model", "llama-text-embed-v2")
    namespace = params.get("namespace", "default")
    score_threshold = params.get("score_threshold", 0.5)

    if not pc.has_index(index_name):
        logger.info("Index %r not found. Creating with managed embeddings...", index_name)
        pc.create_index_for_model(
            name=index_name,
            cloud=cloud,
            region=region,
            embed={
                "model": embed_model,
                # Our upserts will place text in metadata["chunk_text"]
                "field_map": {"text": "chunk_text"},
            },
        )
        logger.info("Index %r created.", index_name)
    else:
        logger.info("Index %r already exists.", index_name)

    samples = [
        # Source city / airport
        {"id": "src1", "chunk_text": "What is the source city?", "value": "New York (JFK)", "label": "source_city"},
        {"id": "src2", "chunk_text": "Where are we departing from?", "value": "New York (JFK)", "label": "source_city"},
        {"id": "src3", "chunk_text": "Departure airport?", "value": "JFK", "label": "source_airport"},

        # Destination city / airport
        {"id": "dst1", "chunk_text": "What is the destination city?", "value": "San Francisco (SFO)", "label": "destination_city"},
        {"id": "dst2", "chunk_text": "Where are we flying to?", "value": "San Francisco (SFO)", "label": "destination_city"},
        {"id": "dst3", "chunk_text": "Arrival airport?", "value": "SFO", "label": "destination_airport"},

        # Dates
        {"id": "dpt1", "chunk_text": "When are we leaving?", "value": "2025-09-12", "label": "depart_date"},
        {"id": "dpt2", "chunk_text": "Departure date", "value": "2025-09-12", "label": "depart_date"},
        {"id": "ret1", "chunk_text": "When do we come back?", "value": "2025-09-19", "label": "return_date"},
        {"id": "ret2", "chunk_text": "Return date", "value": "2025-09-19", "label": "return_date"},
    ]

    index = pc.Index(index_name)
    stats = index.describe_index_stats()
    if stats["total_vector_count"] == 0:
        logger.info("Upserting %d samples…", len(samples))    
        index.upsert_records(records=samples, namespace=namespace)
    else:
        logger.info("Index %r already contains data.", index_name)
        logger.info("Index stats: %r", stats)
        
    tests = [
        "What is the departure airport?",
        "Where are we flying to?",
        "When is the return flight?",
        "When are we leaving?",
        "What is the destination city?",
    ]
    
    resolver = PineconeManagedEmbedResolver(
        index=index_name,
        embed_model=embed_model,
        namespace=namespace,
        score_threshold=score_threshold,
        top_k=5,
    )
        
    for q in tests:
        out = resolver.resolve(ResolveInput(text=q))
        logger.debug("Query: %s → Result: %s", q, out.value)


if __name__ == "__main__":
    main()
